"""
Jac's set class for Jaseci

Adds relevant operators to id_list for operations on sets of nodes and edges
"""
# from jaseci.utils.id_list import id_list
from jaseci.utils.utils import logger
from jaseci.prim.element import Element


class JacSet(list):
    """
    Jac set class for operations in Jac lang
    (keeps append ordering and no dups)
    """

    def __init__(self, in_list=None):
        if in_list:
            for i in in_list:
                if i not in self:
                    self.append(i)

    def append(self, item):
        if not isinstance(item, Element) or not hasattr(item, "anchor_value"):
            logger.error(f"Invalid {type(item)} object {item} to be added to jac_set!")
        elif item not in self:
            list.append(self, item)

    def add_obj(self, item: Element):
        if item not in self:
            self.append(item)

    def obj_list(self):
        return self

    def __lt__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() < other:
                ret.add_obj(i)
        return ret

    def __gt__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() > other:
                ret.add_obj(i)
        return ret

    def __le__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() <= other:
                ret.add_obj(i)
        return ret

    def __ge__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() >= other:
                ret.add_obj(i)
        return ret

    def __eq__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() == other:
                ret.add_obj(i)
        return ret

    def __ne__(self, other):
        """Returns reduced set where anchor value evals to other"""
        ret = JacSet()
        for i in self:
            if i.anchor_value() != other:
                ret.add_obj(i)
        return ret

    def __add__(self, other):
        """Returns new set with operation applied"""
        ret = JacSet()
        for i in self:
            ret.add_obj(i)
        for i in other:
            if i not in ret:
                ret.add_obj(i)
        return ret

    def __sub__(self, other):
        """Returns new set with operation applied"""
        ret = JacSet()
        for i in self:
            if i not in other:
                ret.add_obj(i)
        return ret

    def __mul__(self, other):
        """Returns new set with operation applied, mul is intersection"""
        ret = JacSet()
        for i in self:
            if i in other:
                ret.add_obj(i)
        return ret

    def __truediv__(self, other):
        """Returns new set with operation applied, div is 'outersection'"""
        ret = JacSet()
        for i in self:
            if i not in other:
                ret.add_obj(i)
        for i in other:
            if i not in self:
                ret.add_obj(i)
        return ret



"""
Edge class for Jaseci

Each edge has an id, name, timestamp, the from node at the element of the edge
and the to node it is pointing to.
"""
from jaseci.prim.element import Element
from jaseci.prim.obj_mixins import Anchored
from jaseci.utils.utils import logger
import uuid
import sys


class Edge(Element, Anchored):
    """Edge class for Jaseci"""

    def __init__(self, **kwargs):
        self.from_node_id = None
        self.to_node_id = None
        self.bidirected: bool = False
        Anchored.__init__(self)
        Element.__init__(self, **kwargs)

    def from_node(self):
        """Returns node edge is pointing from"""
        ret = (
            self._h.get_obj(self._m_id, self.from_node_id)
            if self.from_node_id
            else None
        )
        if not ret:
            logger.critical(str(f"{self} disconnected from source node"))
            return None
        else:
            return ret

    def to_node(self):
        """Returns node edge is pointing to"""
        if not self.to_node_id:
            return None
        ret = self._h.get_obj(self._m_id, self.to_node_id) if self.to_node_id else None
        if not ret:
            logger.critical(str(f"{self} disconnected to target node"))
            return None
        else:
            return ret

    def nodes(self):
        """Returns both nodes connected to edge in a list"""
        return [self.to_node(), self.from_node()]

    def opposing_node(self, node_obj):
        """Returns opposite node edge is pointing from node_obj"""
        node_set = [self.to_node_id, self.from_node_id]
        try:
            node_set.remove(node_obj.jid)
            return self._h.get_obj(self._m_id, node_set[0])
        except ValueError:
            logger.critical(str(f"{self} disconnected to node {node_obj}"))
            return None

    def connect(self, source, target, bi_dir=False):
        """
        Connects both ends of the edge
        """
        self.from_node_id = source.jid
        self.to_node_id = target.jid
        source.smart_add_edge(self)
        target.smart_add_edge(self)
        self.set_bidirected(bi_dir)
        self.save()
        return True

    def set_bidirected(self, bidirected: bool):
        """Sets/unsets edge to be bidirected"""
        self.bidirected = bidirected
        self.save()

    def is_bidirected(self):
        """Check if edge is bidirected"""
        return self.bidirected

    def connects(self, source=None, target=None, ignore_direction=False):
        """Test if a node or nodes are connected by edge"""
        if not source and not target:
            return False
        if self.bidirected or ignore_direction:
            if source and source.jid not in [self.from_node_id, self.to_node_id]:
                return False
            if target and target.jid not in [self.from_node_id, self.to_node_id]:
                return False
        else:
            if source and source.jid != self.from_node_id:
                return False
            if target and target.jid != self.to_node_id:
                return False
        return True

    def is_fast(self):
        return sys.getsizeof(self.context) < 2000

    def save(self):
        """
        Write self through hook to persistent storage
        """
        if self.is_fast():
            self._persist = False
        super().save()

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        base = self.from_node()
        target = self.to_node()
        base.smart_remove_edge(self) if base else None
        target.smart_remove_edge(self) if target else None
        super().destroy()

    def dot_str(self, node_map=None, edge_map=None, detailed=False):
        """
        DOT representation
        from_node -> to_node [context_key=contect_value]
        """

        def handle_str(str):
            return str[:32].replace('"', '\\"')

        from_name = (
            uuid.UUID(self.from_node().jid).hex
            if node_map is None
            else node_map.index(self.from_node().jid)
        )
        to_name = (
            uuid.UUID(self.to_node().jid).hex
            if node_map is None
            else node_map.index(self.to_node().jid)
        )
        dstr = f'"n{from_name}" -> "n{to_name}" [ '

        if detailed:
            dstr += f'id="{uuid.UUID(self.jid).hex}", '

        label = ""
        if edge_map:
            label = f"e{edge_map.index(self.jid)}"
        if self.name != "generic":
            label += f":{self.name}"

        dstr += f'label="{label}"'

        if self.bidirected:
            dstr += ', dir="both"'

        edge_dict = self.context
        for i in self.private_values():
            edge_dict.pop(i)

        if edge_dict and detailed:
            for k, v in edge_dict.items():
                if not isinstance(v, str) or v == "":
                    continue
                dstr += f', {k}="{handle_str(v)}"'

        dstr += " ]"

        return dstr + "\n"



"""Built in actions for Jaseci"""
import stripe as s

from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.stripe_svc import StripeService
from datetime import datetime
from jaseci.jsorc.live_actions import jaseci_action


def stripe() -> s:
    return JsOrc.svc("stripe", StripeService).poke()


@jaseci_action()
def create_product(name: str, description: str, **kwargs):
    """create product"""

    return stripe().Product.create(name=name, description=description, **kwargs)


@jaseci_action()
def create_product_price(
    productId: str,
    amount: int,
    currency: str,
    recurring: dict = {},
    **kwargs,
):
    """modify product price"""

    return stripe().Price.create(
        product=productId,
        unit_amount=amount,
        currency=currency,
        **recurring if recurring else {},
        **kwargs,
    )


@jaseci_action()
def product_list(detailed: bool = False, **kwargs):
    """retrieve all producs"""

    return stripe().Product.list(**{"active": True} if detailed else {}, **kwargs)


@jaseci_action()
def create_customer(
    email: str,
    name: str,
    address: dict = {},
    **kwargs,
):
    """create customer"""

    return stripe().Customer.create(email=email, name=name, address=address, **kwargs)


@jaseci_action()
def get_customer(customer_id: str, **kwargs):
    """retrieve customer information"""

    return stripe().Customer.retrieve(id=customer_id, **kwargs)


@jaseci_action()
def attach_payment_method(payment_method_id: str, customer_id: str, **kwargs):
    """attach payment method to customer"""

    paymentMethods = get_payment_methods(customer_id)

    paymentMethod = stripe().PaymentMethod.attach(
        payment_method=payment_method_id, customer=customer_id, **kwargs
    )

    is_default = True
    if paymentMethods.get("data"):
        update_default_payment_method(customer_id, payment_method_id)
        is_default = False

    paymentMethod["is_default"] = is_default

    return paymentMethod


@jaseci_action()
def detach_payment_method(payment_method_id: str, **kwargs):
    """detach payment method from customer"""

    return stripe().PaymentMethod.detach(payment_method=payment_method_id, **kwargs)


@jaseci_action()
def get_payment_methods(customer_id: str, **kwargs):
    """get customer list of payment methods"""

    return stripe().PaymentMethod.list(customer=customer_id, type="card", **kwargs)


@jaseci_action()
def update_default_payment_method(customer_id: str, payment_method_id: str, **kwargs):
    """update default payment method of customer"""

    return stripe().Customer.modify(
        sid=customer_id,
        invoice_settings={"default_payment_method": payment_method_id},
        **kwargs,
    )


@jaseci_action()
def create_invoice(customer_id: str, **kwargs):
    """create customer invoice"""

    return stripe().Invoice.create(customer=customer_id, **kwargs)


@jaseci_action()
def get_invoice_list(
    customer_id: str,
    subscription_id: str,
    starting_after: str = "",
    limit: int = 10,
    **kwargs,
):
    """retrieve customer list of invoices"""

    return stripe().Invoice.list(
        customer=customer_id,
        limit=limit,
        subscription=subscription_id,
        **{"starting_after": starting_after} if starting_after else {},
        **kwargs,
    )


@jaseci_action()
def get_payment_intents(
    customer_id: str, starting_after: str = "", limit: int = 10, **kwargs
):
    """get customer payment intents"""

    return stripe().PaymentIntent.list(
        customer=customer_id,
        limit=limit,
        **{"starting_after": starting_after} if starting_after else {},
        **kwargs,
    )


@jaseci_action()
def create_payment_intents(
    customer_id: str,
    amount: int,
    currency: str,
    payment_method_types: list = ["card"],
    **kwargs,
):
    """Create customer payment"""

    return stripe().PaymentIntent.create(
        customer=customer_id,
        amount=amount,
        currency=currency,
        payment_method_types=payment_method_types,
        **kwargs,
    )


@jaseci_action()
def get_customer_subscription(customer_id: str, **kwargs):
    """retrieve customer subcription list"""

    return stripe().Subscription.list(customer=customer_id, **kwargs)


@jaseci_action()
def create_payment_method(card_type: str, card: dict, billing_details: dict, **kwargs):
    """create payment method"""

    return stripe().PaymentMethod.create(
        type=card_type, card=card, billing_details=billing_details, **kwargs
    )


@jaseci_action()
def create_trial_subscription(
    customer_id: str,
    items: list,
    payment_method_id: str = "",
    trial_period_days: int = 14,
    **kwargs,
):
    """create customer trial subscription"""

    if payment_method_id != "":
        # attach payment method to customer
        attach_payment_method(payment_method_id, customer_id)

        # set card to default payment method
        update_default_payment_method(customer_id, payment_method_id)

    return stripe().Subscription.create(
        customer=customer_id,
        items=items,
        trial_period_days=trial_period_days,
        **kwargs,
    )


@jaseci_action()
def create_subscription(
    customer_id: str,
    items: list,
    payment_method_id: str = "",
    **kwargs,
):
    """create customer subscription"""

    if payment_method_id != "":
        # attach payment method to customer
        attach_payment_method(payment_method_id, customer_id)

        # set card to default payment method
        update_default_payment_method(customer_id, payment_method_id)

    return stripe().Subscription.create(customer=customer_id, items=items, **kwargs)


@jaseci_action()
def cancel_subscription(subscription_id: str, **kwargs):
    """cancel customer subscription"""

    return stripe().Subscription.delete(sid=subscription_id, **kwargs)


@jaseci_action()
def get_subscription(subscription_id: str, **kwargs):
    """retrieve customer subcription details"""

    return stripe().Subscription.retrieve(id=subscription_id, **kwargs)


@jaseci_action()
def update_subscription_item(
    subscription_id: str, subscription_item_id: str, price_id: str, **kwargs
):
    """update subcription details"""

    return stripe().Subscription.modify(
        sid=subscription_id,
        cancel_at_period_end=False,
        items=[
            {
                "id": subscription_item_id,
                "price": price_id,
            },
        ],
        **kwargs,
    )


@jaseci_action()
def get_invoice(invoice_id: str, **kwargs):
    """get invoice information"""

    return stripe().Invoice.retrieve(id=invoice_id, **kwargs)


@jaseci_action()
def create_usage_report(subscription_item_id: str, quantity: int, **kwargs):
    """Create usage record"""

    return stripe().SubscriptionItem.create_usage_record(
        id=subscription_item_id, quantity=quantity, timestamp=datetime.now(), **kwargs
    )


@jaseci_action()
def create_checkout_session(
    success_url: str, cancel_url: str, line_items: list, mode: str, **kwargs
):
    return stripe().checkout.Session.create(
        success_url=success_url,
        cancel_url=cancel_url,
        line_items=line_items,
        mode=mode,
        **kwargs,
    )


@jaseci_action()
def exec(api: str, *args, **kwargs):
    apis = api.split(".")

    if not apis:
        raise Exception("API is required!")

    mod = stripe()

    for api in apis:
        mod = getattr(mod, api, None)
        if not mod:
            raise Exception("Not a valid stripe API!")

    return mod(*args, **kwargs)



from jaseci.jsorc.jsorc import JsOrc
from redis import Redis


@JsOrc.service(
    name="redis",
    config="REDIS_CONFIG",
    manifest="REDIS_MANIFEST",
    priority=0,
    proxy=True,
)
class RedisService(JsOrc.CommonService):
    def run(self):
        self.app = Redis(**self.config, decode_responses=True)
        self.app.ping()

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def get(self, name):
        return self.app.get(name)

    def set(self, name, val):
        self.app.set(name, val)

    def exists(self, name):
        return self.app.exists(name)

    def delete(self, name):
        self.app.delete(name)

    def hget(self, name, key):
        return self.app.hget(name, key)

    def hset(self, name, key, val):
        self.app.hset(name, key, val)

    def hexists(self, name, key):
        return self.app.hexists(name, key)

    def hdel(self, name, key):
        self.app.hdel(name, key)

    def hkeys(self, name):
        return self.app.hkeys(name)

    ###################################################
    #                     CLEANER                     #
    ###################################################

    def clear(self):
        if self.is_running():
            self.app.flushdb()

    # ---------------- PROXY EVENTS ----------------- #

    def on_delete(self):
        if self.is_running():
            self.app.close()



from jaseci.jac.ir.passes.ir_pass import IrPass
from jaseci.jac.jsci_vm.disasm import DisAsm


class PrinterPass(IrPass):
    def __init__(self, to_screen=True, with_exit=False, **kwargs):
        super().__init__(**kwargs)
        self.to_screen = to_screen
        self.with_exit = with_exit
        self.output = []

    def enter_node(self, node):
        out = "Entering " if self.with_exit else "Encountering " + str(node)
        print(out) if self.to_screen else None
        self.output.append(out)
        if hasattr(node, "bytecode"):
            self.output += DisAsm().disassemble(node.bytecode, to_screen=self.to_screen)

    def exit_node(self, node):
        if self.with_exit:
            out = "Exiting " + str(node)
            print(out) if self.to_screen else None
            self.output.append(out)



import json
from json import JSONDecoder, JSONEncoder

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.id_list import IdList
from jaseci.utils.utils import logger


class JaseciJsonEncoder(JSONEncoder):
    def default(self, obj):
        from jaseci.prim.element import Element

        if isinstance(obj, Element):
            return {"__mem_id__": obj.jid}
        else:
            return super().default(obj)


class JaseciJsonDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if isinstance(obj, dict):
            if "__mem_id__" in obj:
                return self.convert(obj["__mem_id__"])
            else:
                for k in obj:
                    self.transform(obj, k)
        elif isinstance(obj, (list, tuple)):
            for idx, k in enumerate(obj):
                self.transform(obj, idx)
        return obj

    def transform(self, obj, key):
        if isinstance(obj[key], dict):
            if "__mem_id__" in obj[key]:
                obj[key] = self.convert(obj[key]["__mem_id__"])
            else:
                for k in obj[key]:
                    self.transform(obj[key], k)
        elif isinstance(obj[key], (list, tuple)):
            for idx, k in enumerate(obj[key]):
                self.transform(obj[key], idx)

    def convert(self, urn):
        return JsOrc.hook().get_obj_from_store(urn)


def json_str_to_jsci_dict(input_str, parent_obj=None):
    """
    Helper function to convert JSON strings to dictionarys with _ids list
    conversions from hex to UUID

    ret_obj is the owning object for id_list objects
    """

    try:
        obj_fields = json.loads(input_str, cls=JaseciJsonDecoder)
    except ValueError:
        logger.error(str(f"Invalid jsci_obj string {input_str} on {parent_obj.jid}"))
        obj_fields = {}
    for i in obj_fields.keys():
        if str(i).endswith("_ids") and isinstance(obj_fields[i], list):
            obj_fields[i] = IdList(parent_obj=parent_obj, in_list=obj_fields[i])
    return obj_fields



import signal
import psycopg2

from time import sleep
from copy import deepcopy
from json import dumps, loads
from datetime import datetime
from typing import TypeVar, Any, Union

from jaseci.utils.utils import logger
from jaseci.jsorc.jsorc_settings import JsOrcSettings
from jaseci.jsorc.jsorc_utils import State, CommonService as cs, ManifestType

from kubernetes.client.rest import ApiException

# For future use
# from concurrent.futures import ThreadPoolExecutor
# from os import cpu_count

T = TypeVar("T")


class JsOrc:
    # ------------------- ALIAS ------------------- #

    CommonService = cs

    # ----------------- REFERENCE ----------------- #

    _contexts = {}
    _context_instances = {}

    _services = {}
    _service_instances = {}

    _repositories = {}

    # ----------------- SETTINGS ----------------- #

    _use_proxy = False
    _settings = JsOrcSettings

    # For future use
    # _executor = ThreadPoolExecutor(
    #     min(4, int((cpu_count() or 2) / 4) + 1)
    # )

    # ------------------ COMMONS ------------------ #

    _backoff_interval = 10
    _running_interval = 0
    __running__ = False
    __proxy__ = cs.proxy()

    # ------------------- REGEN ------------------- #

    _regeneration_queues = []
    _regenerating = False
    _has_db = False

    @staticmethod
    def push(name: str, target: dict, entry: dict):
        if name not in target:
            target[name] = [entry]
        else:
            target[name] = sorted(
                target[name] + [entry],
                key=lambda item: (-item["priority"], -item["date_added"]),
            )

    @classmethod
    def run(cls):
        if not cls.__running__:
            cls.__running__ == True
            config: dict = cls.settings("JSORC_CONFIG")
            if cls.db_check():
                hook = cls.hook()
                config = hook.get_or_create_glob("JSORC_CONFIG", config)
            cls._backoff_interval = max(5, config.get("backoff_interval", 10))
            cls._regeneration_queues = config.get("pre_loaded_services", [])
            cls.push_interval(1)

    @classmethod
    def push_interval(cls, interval):
        if cls._running_interval == 0:
            cls._running_interval += 1
            signal.alarm(interval)
        else:
            logger.info("Reusing current running interval...")

    @classmethod
    def kube(cls):
        if not cls._kube:
            raise Exception(f"Kubernetes is not yet ready!")
        return cls._kube

    #################################################
    #                    HELPER                     #
    #################################################

    # ------------------ context ------------------ #

    @classmethod
    def get(cls, context: str, cast: T = None, *args, **kwargs) -> Union[T, Any]:
        """
        Get existing context instance or build a new one that will persists
        ex: master

        context: name of the context to be build
        cast: to cast the return and allow code hinting
        *arsgs: additional argument used to initialize context
        **kwargs: additional keyword argument used to initialize context
        """
        if context not in cls._contexts:
            raise Exception(f"Context {context} is not existing!")

        if context not in cls._context_instances:
            cls._context_instances[context] = cls.ctx(context, cast, *args, **kwargs)

        return cls._context_instances[context]

    @classmethod
    def destroy(cls, context: str):
        """
        remove existing context instance
        ex: master

        context: name of the context to be build
        """
        if context not in cls._contexts:
            raise Exception(f"Context {context} is not existing!")

        if context in cls._context_instances:
            del cls._context_instances[context]

    @classmethod
    def renew(cls, context: str, cast: T = None, *args, **kwargs) -> Union[T, Any]:
        """
        renew existing context instance or build a new one that will persists
        ex: master

        context: name of the context to be build
        cast: to cast the return and allow code hinting
        *arsgs: additional argument used to initialize context
        **kwargs: additional keyword argument used to initialize context
        """
        cls.destroy(context)

        cls._context_instances[context] = cls.ctx(context, cast, *args, **kwargs)
        return cls._context_instances[context]

    @classmethod
    def ctx(cls, context: str, cast: T = None, *args, **kwargs) -> Union[T, Any]:
        """
        Build new instance of the context
        ex: master

        context: name of the context to be build
        cast: to cast the return and allow code hinting
        *arsgs: additional argument used to initialize context
        **kwargs: additional keyword argument used to initialize context
        """
        if context not in cls._contexts:
            raise Exception(f"Context {context} is not existing!")

        # highest priority
        context = cls._contexts[context][0]

        return context["type"](*args, **kwargs)

    @classmethod
    def ctx_cls(cls, context: str):
        """
        Get the context class
        """
        if context not in cls._contexts:
            return None

        return cls._contexts[context][0]["type"]

    @classmethod
    def master(cls, cast: T = None, *args, **kwargs) -> Union[T, Any]:
        """
        Generate master instance
        """
        return cls.__gen_with_hook("master", cast, *args, **kwargs)

    @classmethod
    def super_master(cls, cast: T = None, *args, **kwargs) -> Union[T, Any]:
        """
        Generate super_master instance
        """
        return cls.__gen_with_hook("super_master", cast, *args, **kwargs)

    @classmethod
    def __gen_with_hook(cls, context: str, *args, **kwargs):
        """
        Common process on master and super_master
        """
        if not kwargs.get("h", None):
            kwargs["h"] = cls.hook()

        return cls.ctx(context, None, *args, **kwargs)

    # ------------------ service ------------------ #

    @classmethod
    def _svc(cls, service: str) -> cs:
        if service not in cls._services:
            raise Exception(f"Service {service} is not existing!")

        # highest priority
        instance = cls._services[service][0]

        config = cls.settings(instance["config"], cls.settings("DEFAULT_CONFIG"))
        manifest = cls.settings(
            instance["manifest"] or "DEFAULT_MANIFEST", cls.settings("DEFAULT_MANIFEST")
        )

        if cls.db_check():
            hook = cls.hook(use_proxy=instance["proxy"])

            config = hook.get_or_create_glob(instance["config"], config)

            manifest = (
                hook.get_or_create_glob(instance["manifest"], manifest)
                if instance["manifest"]
                else {}
            )

        instance: cs = instance["type"](
            config, manifest, instance["manifest_type"], instance
        )

        if instance.has_failed() and service not in cls._regeneration_queues:
            cls._regeneration_queues.append(service)

        return instance

    @classmethod
    def svc(cls, service: str, cast: T = None) -> Union[T, cs]:
        """
        Get service. Initialize when not yet existing.
        ex: task

        service: name of the service to be reference
        cast: to cast the return and allow code hinting
        """
        if service not in cls._services:
            raise Exception(f"Service {service} is not existing!")

        if service not in cls._service_instances:
            if cls._use_proxy and cls._services[service][0]["proxy"]:
                return cls.__proxy__
            cls._service_instances[service] = cls._svc(service)

        return cls._service_instances[service]

    @classmethod
    def svc_reset(cls, service, cast: T = None) -> Union[T, cs]:
        """
        Service reset now deletes the actual instance and rebuild it
        """
        if service in cls._service_instances:
            instance = cls._service_instances.pop(service)
            del instance
        return cls.svc(service)

    # ---------------- repository ----------------- #

    @classmethod
    def src(cls, repository: str, cast: T = None) -> Union[T, Any]:
        """
        Initialize datasource class (repository)
        ex: hook

        repository: name of the repository to be reference
        cast: to cast the return and allow code hinting
        """
        if repository not in cls._repositories:
            raise Exception(f"Repository {repository} is not existing!")

        # highest priority
        repository = cls._repositories[repository][0]

        return repository["type"]()

    @classmethod
    def hook(cls, cast: T = None, use_proxy: bool = False) -> Union[T, Any]:
        """
        Generate hook repository instance
        """
        cls._use_proxy = use_proxy
        hook = cls.src("hook")
        cls._use_proxy = False
        return hook

    #################################################
    #                  DECORATORS                   #
    #################################################

    @classmethod
    def service(
        cls,
        name: str = None,
        config: str = None,
        manifest: str = None,
        manifest_type: ManifestType = ManifestType.DEDICATED,
        priority: int = 0,
        proxy: bool = False,
    ):
        """
        Save the class in services options
        name: name to be used for reference
        config: config name from datasource
        manifest: manifest name from datasource
        priority: duplicate name will use the highest priority
        proxy: allow proxy service
        manifest_type: manifest process type
        """

        def decorator(service: T) -> T:
            cls.push(
                name=name or service.__name__,
                target=cls._services,
                entry={
                    "type": service,
                    "config": config or f"{name.upper()}_CONFIG",
                    "manifest": manifest,
                    "manifest_type": manifest_type,
                    "priority": priority,
                    "proxy": proxy,
                    "date_added": int(datetime.utcnow().timestamp() * 1000),
                },
            )
            return service

        return decorator

    @classmethod
    def repository(cls, name: str = None, priority: int = 0):
        """
        Save the class in repositories options
        name: name to be used for reference
        priority: duplicate name will use the highest priority
        """

        def decorator(repository: T) -> T:
            cls.push(
                name=name or repository.__name__,
                target=cls._repositories,
                entry={
                    "type": repository,
                    "priority": priority,
                    "date_added": int(datetime.utcnow().timestamp() * 1000),
                },
            )
            return repository

        return decorator

    @classmethod
    def context(cls, name: str = None, priority: int = 0):
        """
        Save the class in contexts options
        name: name to be used for reference
        priority: duplicate name will use the highest priority
        """

        def decorator(context: T) -> T:
            cls.push(
                name=name or context.__name__,
                target=cls._contexts,
                entry={
                    "type": context,
                    "priority": priority,
                    "date_added": int(datetime.utcnow().timestamp() * 1000),
                },
            )
            return context

        return decorator

    @classmethod
    def inject(cls, contexts: list = [], services: list = [], repositories: list = []):
        """
        Allow to inject instance on specific method/class
        contexts: list of context name to inject
            - can use tuple per entry (name, alias) instead of string
        services: list of service name to inject
            - can use tuple per entry (name, alias) instead of string
        repositories: list of service name to inject
            - can use tuple per entry (name, alias) instead of string
        """

        def decorator(callable):
            def argument_handler(*args, **kwargs):
                _instances = {}

                for context in contexts:
                    if isinstance(context, tuple):
                        _instances[context[1]] = cls.ctx(context[0])
                    else:
                        _instances[context] = cls.ctx(context)
                for repository in repositories:
                    if isinstance(repository, tuple):
                        _instances[repository[1]] = cls.src(repository[0])
                    else:
                        _instances[repository] = cls.src(repository)
                for service in services:
                    if isinstance(service, tuple):
                        _instances[service[1]] = cls.svc(service[0])
                    else:
                        _instances[service] = cls.svc(service)

                kwargs.update(_instances)
                callable(*args, **kwargs)

            return argument_handler

        return decorator

    @classmethod
    def settings(cls, name: str, default: T = None) -> Union[T, Any]:
        return getattr(cls._settings, name, default)

    @classmethod
    def overrided_namespace(
        cls, name: str, manifest_type: ManifestType = ManifestType.DEDICATED
    ) -> tuple:
        manual_namespace = cls.settings("SERVICE_MANIFEST_MAP").get(name)
        if manual_namespace:
            if manual_namespace == "SOURCE":
                manifest_type = ManifestType.SOURCE
            else:
                manual_namespace == manual_namespace.lower()
                manifest_type = ManifestType.MANUAL
                return manifest_type, manual_namespace

        return (manifest_type,)

    #################################################
    #                  AUTOMATION                   #
    #################################################

    @classmethod
    def add_regeneration_queue(cls, service: str):
        cls.svc(service).state = State.RESTART
        cls._regeneration_queues.append(service)

    @classmethod
    def regenerate(cls):
        if not cls._regenerating:
            cls._regenerating = True

            if cls._has_db:
                cls.regenerate_service()
            else:
                cls.regenerate_database()

            cls._regenerating = False

    @classmethod
    def regenerate_service(cls):
        from jaseci.extens.svc.kube_svc import KubeService
        from jaseci.utils.actions.actions_manager import ActionManager

        kube = cls.svc("kube", KubeService)
        while cls._regeneration_queues:
            regeneration_queue = cls._regeneration_queues.pop(0)
            service = cls.svc(regeneration_queue)
            hook = cls.hook(use_proxy=service.source["proxy"])
            if not service.is_running() and service.enabled and service.automated:
                if service.manifest and kube.is_running():
                    manifest = kube.resolve_manifest(
                        hook.get_or_create_glob(
                            service.source["manifest"], service.manifest
                        ),
                        *cls.overrided_namespace(
                            regeneration_queue, service.manifest_type
                        ),
                    )

                    rmhists: dict = hook.get_or_create_glob(
                        "RESOLVED_MANIFEST_HISTORY", {}
                    )

                    _rmhist = rmhists.get(service.source["manifest"], [{}])[0]
                    rmhist = deepcopy(_rmhist)

                    for kind, confs in manifest.items():
                        for name, conf in confs.items():
                            namespace = conf["metadata"].get("namespace")

                            if kind in rmhist and name in rmhist[kind]:
                                rmhist[kind].pop(name, None)

                            res = kube.read(kind, name, namespace)
                            if hasattr(res, "status") and res.status == 404:
                                kube.create(kind, name, conf, namespace)
                            elif not isinstance(res, ApiException):
                                config_version = 1

                                if isinstance(res, dict):
                                    if "labels" in res["metadata"]:
                                        config_version = (
                                            res["metadata"]
                                            .get("labels", {})
                                            .get("config_version", 1)
                                        )
                                elif res.metadata.labels:
                                    config_version = res.metadata.labels.get(
                                        "config_version", 1
                                    )

                                if config_version != conf.get("metadata").get(
                                    "labels", {}
                                ).get("config_version", 1):
                                    kube.patch(kind, name, conf, namespace)

                    for kind, confs in rmhist.items():
                        for name, conf in confs.items():
                            namespace = conf["metadata"].get("namespace")
                            res = kube.read(kind, name, namespace, quiet=True)
                            if not isinstance(res, ApiException) and (
                                (isinstance(res, dict) and res.get("metadata"))
                                or res.metadata
                            ):
                                if kind not in cls.settings(
                                    "UNSAFE_KINDS"
                                ) or service.manifest_unsafe_paraphrase == cls.settings(
                                    "UNSAFE_PARAPHRASE"
                                ):
                                    kube.delete(kind, name, namespace)
                                else:
                                    logger.info(
                                        f"You don't have permission to delete `{kind}` for `{name}` with namespace `{namespace}`!"
                                    )

                    if _rmhist != manifest:
                        if service.source["manifest"] not in rmhists:
                            rmhists[service.source["manifest"]] = [manifest]
                        else:
                            rmhists[service.source["manifest"]].insert(0, manifest)
                        hook.save_glob("RESOLVED_MANIFEST_HISTORY", dumps(rmhists))
                        hook.commit()

                cls.svc_reset(regeneration_queue)
            sleep(1)

        action_manager = cls.get("action_manager", ActionManager)
        action_manager.optimize(jsorc_interval=cls._backoff_interval)
        action_manager.record_system_state()

    @classmethod
    def regenerate_database(cls):
        from jaseci.extens.svc.kube_svc import KubeService

        kube = cls.svc("kube", KubeService)

        if kube.is_running():
            while not cls.db_check():
                for kind, confs in kube.resolve_manifest(
                    cls.settings("DB_REGEN_MANIFEST", {}),
                    *cls.overrided_namespace("database"),
                ).items():
                    for name, conf in confs.items():
                        namespace = conf["metadata"].get("namespace")
                        res = kube.read(kind, name, namespace)
                        if hasattr(res, "status") and res.status == 404 and conf:
                            kube.create(kind, name, conf, namespace)
                sleep(1)
        raise SystemExit("Force termination to restart the pod!")

    @classmethod
    def db_check(cls):
        if not cls._has_db:
            try:
                dbrc = cls.settings("DB_REGEN_CONFIG")
                if dbrc["enabled"]:
                    connection = psycopg2.connect(
                        host=dbrc["host"],
                        dbname=dbrc["db"],
                        user=dbrc["user"],
                        password=dbrc["password"],
                        port=dbrc["port"],
                    )
                    connection.close()
                cls._has_db = True
            except Exception:
                cls._has_db = False
        return cls._has_db


def interval_check(signum, frame):
    JsOrc.regenerate()

    # wait interval_check to be finished before decrement
    JsOrc._running_interval -= 1
    JsOrc.push_interval(JsOrc._backoff_interval)


signal.signal(signal.SIGALRM, interval_check)



import re
import os
from enum import Enum
from yaml import safe_load_all, YAMLError
from typing import TypeVar, Any, Union

from multiprocessing import Process, current_process

from jaseci.utils.utils import logger

#######################################################################################################
#                                               HELPERS                                               #
#######################################################################################################

T = TypeVar("T")


def convert_yaml_manifest(file):
    manifest = {}
    try:
        for conf in safe_load_all(file):
            kind = conf["kind"]
            if not manifest.get(kind):
                manifest[kind] = {}
            manifest[kind].update({conf["metadata"]["name"]: conf})
    except YAMLError as exc:
        manifest = {"error": f"{exc}"}

    return manifest


def load_default_yaml(file):
    manifest = {}
    with open(
        f"{os.path.dirname(os.path.abspath(__file__))}/manifests/{file}.yaml", "r"
    ) as stream:
        manifest = convert_yaml_manifest(stream)

    return manifest


def get_service_map(*services):
    service_map = {}
    for service in services:
        namespace = os.getenv(f"{service.upper()}_NAMESPACE")
        if namespace:
            service_map[service] = namespace

    return service_map


#######################################################################################################
#                                              RESOLVER                                               #
#######################################################################################################

placeholder_full = re.compile(r"^\$j\{(.*?)\}$")
placeholder_partial = re.compile(r"\$j\{(.*?)\}")
placeholder_splitter = re.compile(r"\.?([^\.\[\"\]]+)(?:\[\"?([^\"\]]+)\"?\])?")
# original
# placeholder_splitter = re.compile(r"([^\.\[\"\]]+)(?:\[\"?([^\"\]]+)\"?\])?(?:\.([^\.\[\"\]]+))?")


def get_splitter(val: str):
    matches = placeholder_splitter.findall(val)
    _matches = []
    for match in matches:
        for m in match:
            if m:
                _matches.append(m)
    return _matches


def get_value(source: dict, keys: list):
    if keys:
        key = keys.pop(0)
        if key in source:
            if keys:
                if isinstance(source[key], dict):
                    return get_value(source[key], keys)
            else:
                return source[key]
    return None


def placeholder_resolver(manifest, data: dict or list):
    for k, d in data.items() if isinstance(data, dict) else enumerate(data):
        if isinstance(d, (dict, list)):
            placeholder_resolver(manifest, d)
        elif isinstance(d, str):
            matcher = placeholder_full.search(d)
            if matcher:
                keys = get_splitter(matcher.group(1))
                data[k] = get_value(manifest, keys)
            else:
                for matcher in placeholder_partial.findall(d):
                    keys = get_splitter(matcher)
                    data[k] = data[k].replace(
                        "$j{" + matcher + "}", get_value(manifest, keys)
                    )


#######################################################################################################
#                                                ENUMS                                                #
#######################################################################################################


class State(Enum):
    RESTART = -2
    FAILED = -1
    NOT_STARTED = 0
    STARTED = 1
    RUNNING = 2

    def is_ready(self):
        return self == State.NOT_STARTED

    def is_running(self):
        return self == State.RUNNING

    def has_failed(self):
        return self == State.FAILED


class ManifestType(Enum):
    SOURCE = -1
    DEDICATED = 0
    MANUAL = 1


#######################################################################################################
#                                               COMMON                                                #
#######################################################################################################


class CommonService:
    ###################################################
    #                   PROPERTIES                    #
    ###################################################

    # ------------------- DAEMON -------------------- #

    _daemon = {}

    @property
    def daemon(self):
        return __class__._daemon

    ###################################################
    #                   INITIALIZER                   #
    ###################################################

    def __init__(
        self,
        config: dict,
        manifest: dict,
        manifest_type: ManifestType = ManifestType.DEDICATED,
        source: dict = {},
    ):
        self.app = None
        self.error = None
        self.state = State.NOT_STARTED
        self.source = source

        # ------------------- CONFIG -------------------- #

        self.config = config
        self.enabled = config.pop("enabled", False)
        self.quiet = config.pop("quiet", False)
        self.automated = config.pop("automated", False)

        # ------------------ MANIFEST ------------------- #

        self.manifest = manifest
        self.manifest_type = manifest_type
        self.manifest_unsafe_paraphrase = manifest.pop("__UNSAFE_PARAPHRASE__", "")

        self.start()

    ###################################################
    #                     BUILDER                     #
    ###################################################

    def start(self):
        try:
            if self.enabled and self.is_ready():
                self.state = State.STARTED
                self.run()
                self.state = State.RUNNING
                self.post_run()
        except Exception as e:
            if not (self.quiet):
                logger.error(
                    f"Skipping {self.__class__.__name__} due to initialization "
                    f"failure!\n{e.__class__.__name__}: {e}"
                )
            self.failed(e)

        return self

    def run(self):
        raise Exception(f"Not properly configured! Please override run method!")

    def post_run(self):
        pass

    # ------------------- DAEMON -------------------- #

    def spawn_daemon(self, **targets):
        if current_process().name == "MainProcess":
            for name, target in targets.items():
                dae: Process = self.daemon.get(name)
                if not dae or not dae.is_alive():
                    process = Process(target=target, daemon=True)
                    process.start()
                    self.daemon[name] = process

    def terminate_daemon(self, *names):
        for name in names:
            dae: Process = self.daemon.pop(name, None)
            if not (dae is None) and dae.is_alive():
                logger.info(f"Terminating {name} ...")
                dae.terminate()

    ###################################################
    #                     COMMONS                     #
    ###################################################

    def poke(self, cast: T = None, msg: str = None) -> Union[T, Any]:
        if self.is_running():
            return (
                self if cast and cast.__name__ == self.__class__.__name__ else self.app
            )
        raise Exception(
            msg or f"{self.__class__.__name__} is disabled or not yet configured!"
        )

    def is_ready(self):
        return self.state.is_ready() and self.app is None

    def is_running(self):
        return self.state.is_running() and not (self.app is None)

    def has_failed(self):
        return self.state.has_failed()

    def failed(self, error: Exception = None):
        self.app = None
        self.state = State.FAILED
        self.error = error

    @classmethod
    def proxy(cls):
        return cls({}, {})

    # ---------------- PROXY EVENTS ----------------- #

    def on_delete(self):
        pass

    # ------------------- EVENTS -------------------- #

    def __del__(self):
        self.on_delete()

    def __getstate__(self):
        return {}

    def __setstate__(self, ignored):
        # for build on pickle load
        self.state = State.FAILED
        del self



import re

from jaseci.jsorc.live_actions import jaseci_action


def compile(pattern: str, flags=0):
    return re.compile(pattern=pattern, flags=flags)


@jaseci_action()
def findall(pattern: str, string: str, flags=0):
    return re.findall(pattern=pattern, string=string, flags=flags)


@jaseci_action()
def search(pattern: str, string: str, flags=0):
    m = re.search(pattern=pattern, string=string, flags=flags)
    if m:
        span = m.span()
        match = m.group()
        return {"span": span, "match": match}
    else:
        return None


@jaseci_action()
def match(pattern: str, string: str, flags=0):
    m = re.match(pattern=pattern, string=string, flags=flags)
    if m:
        span = m.span()
        match = m.group()
        return {"span": span, "match": match}
    else:
        return None


@jaseci_action()
def fullmatch(pattern: str, string: str, flags=0):
    m = re.fullmatch(pattern=pattern, string=string, flags=flags)
    if m:
        span = m.span()
        match = m.group()
        return {"span": span, "match": match}
    else:
        return None


@jaseci_action()
def split(pattern: str, string: str, maxsplit=0, flags=0):
    return re.split(pattern=pattern, string=string, maxsplit=maxsplit, flags=flags)


@jaseci_action()
def finditer(pattern: str, string: str, flags=0):
    iter = re.finditer(pattern, string, flags=0)
    ret_list = []
    for item in iter:
        dicta = {"span": item.span(), "match": item.group()}
        ret_list.append(dicta)

    return ret_list


@jaseci_action()
def sub(pattern: str, replace: str, string: str, count=0, flags=0):
    return re.sub(
        pattern=pattern, repl=replace, string=string, count=count, flags=flags
    )


@jaseci_action()
def subn(pattern: str, replace: str, string: str, count=0, flags=0):
    return list(
        re.subn(pattern=pattern, repl=replace, string=string, count=count, flags=flags)
    )


@jaseci_action()
def escape(pattern: str):
    return re.escape(pattern=pattern)


@jaseci_action()
def purge():
    return re.purge()



from unittest import TestCase

from jaseci.prim.architype import Architype
from jaseci.prim import action
from jaseci.prim.edge import Edge
from jaseci.prim.node import Node
from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.utils import TestCaseHelper


class NodeTests(TestCaseHelper, TestCase):
    """Tests for the funcationality of Jaseci node class"""

    def test_node_connections(self):
        """Test connecting and disconnecting etc of nodes"""
        node1 = Node(m_id=0, h=JsOrc.hook())
        node2 = Node(m_id=0, h=node1._h)
        node3 = Node(m_id=0, h=node1._h)
        node4 = Node(m_id=0, h=node1._h)
        node1.attach_outbound(node2)
        node3.attach_outbound(node2)
        node1.attach_inbound(node2)
        node3.attach_inbound(node2)
        self.assertEqual(len(node2.inbound_nodes()), 2)
        self.assertEqual(node2.inbound_nodes()[0], node1)
        self.assertTrue(node2.is_attached_out(node3))
        self.assertTrue(node2.is_attached_in(node1))
        self.assertFalse(node4.is_equivalent(node2))

        node2.detach_inbound(node1)
        node2.detach_inbound(node3)
        node2.detach_outbound(node1)
        node2.detach_outbound(node3)

        self.assertTrue(node4.is_equivalent(node2))
        self.assertTrue(node3.is_equivalent(node1))
        self.assertTrue(node3.is_equivalent(node2))
        self.assertEqual(len(node1.inbound_nodes()), 0)
        self.assertEqual(len(node2.inbound_nodes()), 0)
        self.assertEqual(len(node3.inbound_nodes()), 0)
        self.assertEqual(len(node4.inbound_nodes()), 0)

    def test_add_context_to_node_and_destroy(self):
        """Test adding and removing contexts nodes"""
        node1 = Node(m_id=0, h=JsOrc.hook())
        node1.context["yeah dude"] = "SUP"
        self.assertEqual(node1.context["yeah dude"], "SUP")
        self.assertEqual(len(node1.context.keys()), 1)
        self.assertTrue("yeah dude" in node1.context.keys())
        self.assertFalse("yeah  dude" in node1.context.keys())

    def test_add_entry_action_to_node_and_destroy(self):
        """Test connecting and disconnecting etc of nodes"""
        node1 = Architype(m_id=0, h=JsOrc.hook())
        act = action.Action(m_id=0, h=node1._h, name="yeah dude", value="SUP")
        node1.entry_action_ids.add_obj(act)
        self.assertEqual(
            node1.entry_action_ids.get_obj_by_name("yeah dude").value, "SUP"
        )
        self.assertEqual(len(node1.entry_action_ids), 1)
        self.assertTrue(node1.entry_action_ids.has_obj_by_name("yeah dude"))
        self.assertFalse(node1.entry_action_ids.has_obj_by_name("yeah  dude"))

        node1.entry_action_ids.destroy_obj_by_name(name="yeah dude")
        self.assertEqual(len(node1.entry_action_ids), 0)
        self.assertFalse(node1.entry_action_ids.has_obj_by_name("yeah dude"))

    def test_adding_and_removing_from_hdnodes(self):
        """Test adding nodes and removing them from HDGDs"""
        node1 = Node(m_id=0, h=JsOrc.hook())
        hdgd1 = Node(m_id=0, h=node1._h, name="yeah dude", dimension=1)
        node1.make_member_of(hdgd1)
        self.assertEqual(node1.parent_node_ids.obj_list()[0], hdgd1)
        self.assertEqual(hdgd1.member_node_ids.obj_list()[0], node1)

        hdgd2 = Node(m_id=0, h=node1._h, dimension=2)
        node1.make_member_of(hdgd2)
        self.assertNotIn(hdgd2.id, node1.parent_node_ids)

        node1.leave_memebership_of(hdgd1)
        self.assertEqual(len(hdgd1.member_node_ids), 0)

    def test_inherit_from_element_edge(self):
        """Test that inheriting params with kwargs works"""
        hook = JsOrc.hook()
        a = Edge(
            name="my edge",
            m_id=0,
            h=hook,
        )
        a.connect(Node(m_id=0, h=hook), Node(m_id=0, h=hook)),
        self.assertEqual(a.name, "my edge")



"""
Super master handler for each user of Jaseci, serves as main interface between
between user and Jaseci
"""

from jaseci.extens.api.jsorc_api import JsOrcApi
from jaseci.prim.master import Master
from jaseci.extens.api.logger_api import LoggerApi
from jaseci.extens.api.config_api import ConfigApi
from jaseci.extens.api.global_api import GlobalApi
from jaseci.extens.api.super_api import SuperApi
from jaseci.extens.api.actions_api import ActionsApi
from jaseci.extens.api.prometheus_api import PrometheusApi
from jaseci.jsorc.jsorc import JsOrc


@JsOrc.context(name="super_master")
class SuperMaster(
    Master,
    LoggerApi,
    ActionsApi,
    ConfigApi,
    GlobalApi,
    SuperApi,
    JsOrcApi,
    PrometheusApi,
):
    """Master with admin APIs"""

    def __init__(self, *args, **kwargs):
        Master.__init__(self, *args, **kwargs)



import time
from prometheus_api_client import PrometheusConnect
from jaseci.jsorc.jsorc import JsOrc


class MetricValue:
    node = None
    pod = None

    def __init__(self, _node: dict, _pod: dict):
        self.node = _node
        self.pod = _pod


@JsOrc.service(name="prome", config="PROME_CONFIG", manifest="PROME_MANIFEST")
class PrometheusService(JsOrc.CommonService):
    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self):
        self.app = PrometheusConnect(url=self.config.get("url"), disable_ssl=True)
        self.ping()
        self.cpu = Cpu(self.app)
        self.memory = Memory(self.app)
        self.network = Network(self.app)
        self.disk = Disk(self.app)

    ###################################################
    #                  COMMON UTILS                   #
    ###################################################

    def ping(self) -> bool:
        prom = self.app
        response = prom._session.get(
            "{0}/".format(prom.url),
            verify=prom.ssl_verification,
            headers=prom.headers,
            timeout=2,
        )
        return response.ok

    def all_metrics(self) -> list:
        return self.app.all_metrics()

    def pods(self, namespace: str = "", exclude_prom: bool = False) -> dict:
        if namespace == "":
            util = self.app.get_current_metric_value("kube_pod_info")
        else:
            util = self.app.get_current_metric_value(
                f"kube_pod_info{{namespace='{namespace}'}}"
            )
        res = {}
        for pod in util:
            info = pod["metric"]
            node = info["node"]
            pod = info["pod"]
            if exclude_prom and "prometheus" in pod:
                continue
            if res.get(node) is None:
                res[node] = []
            res[node].append(pod)
        return res

    def node_info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ) -> dict:
        if namespace == "":
            util = self.app.get_current_metric_value("kube_node_info")
        else:
            util = self.app.get_current_metric_value(
                f"kube_node_info{{namespace='{namespace}'}}"
            )

        res = {}
        disk_read = self.disk.read()
        disk_write = self.disk.write()

        node_names = [nodes["metric"]["node"] for nodes in util]

        for node_name in node_names:
            res[node_name] = {}

        for node_name in node_names:
            res[node_name]["disk_read_bytes"] = disk_read.get(node_name, 0)

        for node_name in node_names:
            res[node_name]["disk_write_bytes"] = disk_write.get(node_name, 0)

        return res

    def pod_info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ) -> dict:
        if namespace == "":
            util = self.app.get_current_metric_value("kube_pod_info")
        else:
            util = self.app.get_current_metric_value(
                f"kube_pod_info{{namespace='{namespace}'}}"
            )

        res = {}

        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue

        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue
            # res[pod_name] = pod["metric"]
            res[pod_name] = {}

        if timestamp != 0 and duration != 0:
            cpu = self.cpu.utilization_per_pod_cores(ts=timestamp, duration=duration)
        else:
            cpu = self.cpu.utilization_per_pod_cores()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue
            pod_cpu = cpu.get(pod_name, 0)

            res[pod_name]["cpu_utilization_cores"] = pod_cpu
        if timestamp != 0 and duration != 0:
            cpu = self.cpu.utilization_per_pod_cores(ts=timestamp, duration=duration)
        else:
            cpu = self.cpu.utilization_per_pod_cores()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue
            pod_cpu = cpu.get(pod_name, 0)

            res[pod_name]["cpu_utilization_cores"] = pod_cpu

        mem = self.memory.utilization_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            pod_mem = mem.get(pod_name, 0)
            res[pod_name]["mem_utilization_bytes"] = pod_mem

        if timestamp != 0 and duration != 0:
            recv = self.network.receive_per_pod_bytes(ts=timestamp, duration=duration)
        else:
            recv = self.network.receive_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue
            pod_recv = recv.get(pod_name, 0)
            res[pod_name]["network_recv_bytes"] = pod_recv

        if timestamp != 0 and duration != 0:
            tran = self.network.transmit_per_pod_bytes(ts=timestamp, duration=duration)
        else:
            tran = self.network.transmit_per_pod_bytes()
        for pod in util:
            pod_name = pod["metric"]["pod"]
            if exclude_prom and "prometheus" in pod_name:
                continue
            pod_tran = tran.get(pod_name, 0)
            res[pod_name]["network_tran_bytes"] = pod_tran

        return res

    def info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ) -> MetricValue:
        node_value = self.node_info(
            namespace=namespace,
            exclude_prom=exclude_prom,
            timestamp=timestamp,
            duration=duration,
        )
        pod_value = self.pod_info(
            namespace=namespace,
            exclude_prom=exclude_prom,
            timestamp=timestamp,
            duration=duration,
        )
        return MetricValue(_node=node_value, _pod=pod_value)


class Info:
    def __init__(self, app):
        self.app = app


class Cpu(Info):
    def utilization_core(self) -> dict:
        util = self.app.get_current_metric_value(
            'sum(irate(node_cpu_seconds_total{mode!="idle"}[10m])) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def utilization_percentage(self) -> dict:
        util = self.app.get_current_metric_value(
            '(sum(irate(node_cpu_seconds_total{mode!="idle"}[10m])) by '
            '(node)) / (sum(irate(node_cpu_seconds_total{mode!=""}[10m])) by '
            "(node)) * 100"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def utilization_per_pod_cores(self, ts=int(time.time()), duration=10) -> dict:
        query_str = f'sum(rate(container_cpu_usage_seconds_total{{pod!=""}}[{duration}s] @ {ts})) by (pod)'
        util = self.app.get_current_metric_value(query_str)
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = float(pod["value"][1])
            res[pod_name] = float(value)
        return res


class Memory(Info):
    def total_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum(node_memory_MemTotal_bytes) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def utilization_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum(node_memory_Active_bytes) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def utilization_percentage(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum(node_memory_Active_bytes / node_memory_MemTotal_bytes * 100 ) by "
            "(node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def utilization_per_pod_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            'sum(container_memory_working_set_bytes{pod!=""}) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = float(pod["value"][1])
            res[pod_name] = float(value)
        return res


class Network(Info):
    def receive_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum (rate (node_network_receive_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def receive_per_pod_bytes(self, ts: int = int(time.time()), duration: int = 10):
        util = self.app.get_current_metric_value(
            f'sum (rate (container_network_receive_bytes_total{{pod!=""}}[{duration}s] @ {ts})) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = pod["value"][1]
            res[pod_name] = float(value)
        return res

    def transmit_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum (rate (node_network_transmit_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def transmit_per_pod_bytes(self, ts: int = int(time.time()), duration: int = 10):
        util = self.app.get_current_metric_value(
            f'sum (rate (container_network_transmit_bytes_total{{pod!=""}}[{duration}s] @ {ts})) by (pod)'
        )
        res = {}
        for pod in util:
            pod_name = pod["metric"]["pod"]
            value = pod["value"][1]
            res[pod_name] = float(value)
        return res


class Disk(Info):
    def total_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            'sum(avg (node_filesystem_size_bytes{mountpoint!="/boot", '
            'fstype!="tmpfs"}) without (mountpoint)) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def free_bytes(self) -> dict:
        util = self.app.get_current_metric_value(
            'sum(avg (node_filesystem_free_bytes{mountpoint!="/boot", '
            'fstype!="tmpfs"}) without (mountpoint)) by (node)'
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def read(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum (rate (node_disk_read_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res

    def write(self) -> dict:
        util = self.app.get_current_metric_value(
            "sum (rate (node_disk_written_bytes_total{}[10m])) by (node)"
        )
        res = {}
        for node in util:
            node_name = node["metric"]["node"]
            node_util = float(node["value"][1])
            res[node_name] = node_util
        return res



"""
Sentinel interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from jaseci.prim.node import Node
from jaseci.prim.edge import Edge
from jaseci.prim.walker import Walker
from jaseci.jac.interpreter.interp import Interp
from jaseci.jac.machine.jac_scope import JacScope


class ArchitypeInterp(Interp):
    """Jac interpreter mixin for objects that will execute Jac code"""

    def run_architype(self, jac_ast):
        """
        architype:
            KW_NODE NAME (COLON NAME)* attr_block
            | KW_EDGE NAME (COLON NAME)* attr_block
            | KW_TYPE NAME struct_block
            | KW_GRAPH NAME graph_block
            | KW_ASYNC? KW_WALKER NAME namespaces? walker_block;
        """
        if jac_ast is None:  # Using defaults
            if self.kind == "node" and self.name in ["root", "generic"]:
                return Node(
                    m_id=self._m_id,
                    h=self._h,
                    kind=self.kind,
                    name=self.name,
                    parent=self.parent(),
                )
            elif self.kind == "edge" and self.name in ["generic"]:
                return Edge(
                    m_id=self._m_id,
                    h=self._h,
                    kind=self.kind,
                    name=self.name,
                    parent=self.parent(),
                )

        kid = self.set_cur_ast(jac_ast)

        self.push_scope(JacScope(parent=self, has_obj=self, action_sets=[]))
        if kid[0].name == "KW_NODE":
            item = Node(
                m_id=self._m_id,
                h=self._h,
                kind=kid[0].token_text(),
                name=kid[1].token_text(),
                parent=self.parent(),
            )
            self.build_object_with_supers(item, kid[-1])
        elif kid[0].name == "KW_EDGE":
            item = Edge(
                m_id=self._m_id,
                h=self._h,
                kind=kid[0].token_text(),
                name=kid[1].token_text(),
                parent=self.parent(),
            )
            self.build_object_with_supers(item, kid[-1])
        elif kid[0].name == "KW_TYPE":
            item = self.run_struct_block(kid[-1])
        elif kid[0].name == "KW_GRAPH":
            item = self.run_graph_block(kid[-1])
        elif kid[0].name == "KW_WALKER":
            item = Walker(
                m_id=self._m_id,
                h=self._h,
                code_ir=jac_ast,
                name=kid[1].token_text(),
                kind=kid[0].token_text(),
                parent=self.parent(),
                is_async=self.is_async,
            )
            if kid[2].name == "namespaces":
                item.namespaces = self.run_namespaces(jac_ast.kid[2])
        elif jac_ast.name == "graph_block":  # usedi n jac tests
            item = self.run_graph_block(jac_ast)
        self.pop_scope()
        return item

    def run_namespaces(self, jac_ast):
        """
        namespaces: COLON name_list;
        """
        return self.run_name_list(jac_ast.kid[1])

    def run_attr_block(self, jac_ast, obj):
        """
        attr_block:
            LBRACE (attr_stmt)* RBRACE
            | COLON (attr_stmt)* SEMI
            | SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "attr_stmt":
                self.run_attr_stmt(i, obj)

    def run_struct_block(self, jac_ast):
        """
        struct_block: LBRACE (has_stmt)* RBRACE | COLON has_stmt | SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        ret = {}
        for i in kid:
            if i.name == "has_stmt":
                self.run_has_stmt(i, ret)
        return ret

    def run_can_block(self, jac_ast):
        """
        can_block: (can_stmt)*;
        """
        kid = self.set_cur_ast(jac_ast)
        for i in kid:
            if i.name == "can_stmt":
                self.run_can_stmt(i, self)

    def run_graph_block(self, jac_ast):
        """
        graph_block: graph_block_spawn;
        """
        kid = self.set_cur_ast(jac_ast)
        return getattr(self, f"run_{kid[0].name}")(kid[0])

    def run_graph_block_spawn(self, jac_ast):
        """
        graph_block_spawn:
            LBRACE has_root can_block KW_SPAWN code_block RBRACE
            | COLON has_root can_block KW_SPAWN code_block SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        root_name = self.run_has_root(kid[1])
        self.run_can_block(kid[2])
        m = Interp(parent_override=self.parent(), caller=self)
        m.push_scope(
            JacScope(parent=self, has_obj=None, action_sets=[self.activity_action_ids])
        )
        try:
            m.run_code_block(kid[4])
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", m._cur_jac_ast)
        local_state = m._jac_scope.local_scope
        self.report = self.report + m.report
        if root_name in local_state.keys():
            obj = local_state[root_name]
            if not isinstance(obj, Node):
                self.rt_error(f"{root_name} is {type(obj)} not node!", kid[3])
            return obj
        else:
            self.rt_error("Graph didn't produce root node!", kid[3])
            return None

    def run_has_root(self, jac_ast):
        """
        has_root: KW_HAS KW_ANCHOR NAME SEMI;
        """
        kid = self.set_cur_ast(jac_ast)
        return kid[2].token_text()

    # Helper Functions ##################

    def build_object_with_supers(self, item, jac_ast):
        for i in self.super_archs:
            super_jac_ast = (
                self.parent()
                .arch_ids.get_obj_by_name(name=i, kind=item.kind)
                .get_jac_ast()
                .kid[-1]
            )
            self.run_attr_block(super_jac_ast, item)
        self.run_attr_block(jac_ast, item)



from jaseci.utils.test_core import CoreTest
import uuid


class StackTests(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_walker_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("ll.jac")}],
        )
        self.call(self.mast, ["walker_run", {"name": "init"}])
        self.call(self.mast, ["walker_run", {"name": "gen_rand_life"}])
        ret = self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.assertGreater(len(ret["report"]), 3)
        self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.mast._h.commit()
        self.call(self.mast, ["walker_run", {"name": "get_gen_day"}])
        self.assertEqual(len(self.mast._h.save_obj_list), 0)

    def test_walker_context_auto_refresh(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("ll.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "print_life_note"}])
        life_node = self.mast._h.get_obj(self.mast._m_id, ret["final_node"])
        life_node.context.pop("note")
        ret = self.call(self.mast, ["walker_run", {"name": "print_life_note"}])
        self.assertTrue(ret["success"])

    def test_action_module_list(self):
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertIn("jaseci.extens.act_lib.rand", ret)
        self.assertIn("jaseci.extens.act_lib.std", ret)
        self.assertIn("jaseci.extens.act_lib.file", ret)

    def test_action_module_unload_reload(self):
        ret = self.call(self.smast, ["actions_module_list", {}])
        before = len(ret)
        ret = self.call(
            self.smast,
            ["actions_unload_module", {"name": "jaseci.extens.act_lib.rand"}],
        )
        ret = self.call(
            self.smast,
            ["actions_unload_module", {"name": "jaseci.extens.act_lib.file"}],
        )
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before - 2)
        ret = self.call(
            self.smast,
            ["actions_load_module", {"mod": "jaseci.extens.act_lib.rand"}],
        )
        ret = self.call(
            self.smast,
            ["actions_load_module", {"mod": "jaseci.extens.act_lib.file"}],
        )
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before)

    def test_action_module_unload_reload_aliased(self):
        ret = self.call(self.smast, ["actions_module_list", {}])
        before = len(ret)
        ret = self.call(
            self.smast,
            ["actions_unload_module", {"name": "jaseci.extens.act_lib.vector"}],
        )
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before - 1)
        self.assertNotIn("vector.cos_sim", self.call(self.smast, ["actions_list", {}]))
        ret = self.call(
            self.smast,
            ["actions_load_module", {"mod": "jaseci.extens.act_lib.vector"}],
        )
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before)

    def test_action_unload(self):
        ret = self.call(self.smast, ["actions_module_list", {}])
        before = len(ret)
        ret = self.call(self.smast, ["actions_list", {"name": "rand"}])
        for i in ret:
            self.call(self.smast, ["actions_unload_action", {"name": i}])
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before - 1)
        ret = self.call(
            self.smast,
            ["actions_load_module", {"mod": "jaseci.extens.act_lib.rand"}],
        )

    def test_action_set_unload(self):
        ret = self.call(self.smast, ["actions_module_list", {}])
        before = len(ret)
        ret = self.call(self.smast, ["actions_unload_actionset", {"name": "rand"}])
        ret = self.call(self.smast, ["actions_module_list", {}])
        self.assertEqual(len(ret), before - 1)
        ret = self.call(
            self.smast,
            ["actions_load_module", {"mod": "jaseci.extens.act_lib.rand"}],
        )

    def test_sentinel_missing_architype(self):
        """
        Test when the original sentinel is missing the corresponding architype for a
        node
        """
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("simple.jac")}],
        )
        old_snt = ret[0]["jid"]
        ret = self.call(self.mast, ["walker_run", {"name": "init", "snt": old_snt}])
        node_id = ret["report"][0]["jid"]
        ret = self.call(self.mast, ["sentinel_delete", {"snt": old_snt}])
        ret = self.call(self.mast, ["sentinel_list", {"snt": old_snt}])
        ret = self.call(
            self.mast,
            ["graph_node_set", {"nd": node_id, "ctx": {"b": 6}}],
        )
        self.assertIn("has_var", ret["errors"][0])
        ret = self.call(
            self.mast,
            [
                "sentinel_register",
                {
                    "name": "new_snt",
                    "auto_create_graph": False,
                    "auto_run": False,
                    "set_active": True,
                    "code": self.load_jac("simple.jac"),
                },
            ],
        )
        self.mast._h._machine = None
        ret = self.call(
            self.mast,
            ["graph_node_set", {"nd": node_id, "ctx": {"b": 6}}],
        )
        self.assertEqual(ret["context"]["b"], 6)

    def test_sentinel_missing_architype_global(self):
        """
        Test when the original sentinel is missing the corresponding architype for a
        node
        """
        ret = self.call(
            self.smast,
            [
                "sentinel_register",
                {
                    "name": "new_snt",
                    "auto_create_graph": False,
                    "auto_run": False,
                    "set_active": True,
                    "code": self.load_jac("simple.jac"),
                },
            ],
        )
        glob_snt = ret[0]["jid"]
        ret = self.call(self.smast, ["global_sentinel_set", {"snt": glob_snt}])
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("simple.jac")}],
        )
        old_snt = ret[0]["jid"]
        ret = self.call(self.mast, ["walker_run", {"name": "init", "snt": old_snt}])
        node_id = ret["report"][0]["jid"]
        ret = self.call(self.mast, ["sentinel_delete", {"snt": old_snt}])
        ret = self.call(self.mast, ["sentinel_list", {"snt": old_snt}])
        ret = self.call(
            self.mast,
            ["graph_node_set", {"nd": node_id, "ctx": {"b": 6}}],
        )
        self.assertIn("has_var", ret["errors"][0])
        ret = self.call(self.mast, ["sentinel_active_global", {}])
        self.mast._h._machine = None
        ret = self.call(
            self.mast,
            ["graph_node_set", {"nd": node_id, "ctx": {"b": 6}}],
        )
        self.assertEqual(ret["context"]["b"], 6)

    def test_interp_deep_except(self):
        ret = self.call(
            self.smast,
            [
                "sentinel_register",
                {"auto_run": "", "code": self.load_jac("simple.jac")},
            ],
        )

        ret = self.call(self.smast, ["walker_run", {"name": "deep_except"}])
        self.assertTrue(ret["success"])
        self.assertIn("xxxx.xxxx", ret["report"][2]["msg"])



from io import StringIO
import re
from typing import List, Dict


def match_date(line: str) -> str:
    """Extracts the date from a log line"""
    matched = re.match(r"\d\d\d\d-\d\d-\d\d\ \d\d:\d\d:\d\d", line)
    if matched:
        return matched.group()

    return None


def match_level(line: str) -> str:
    """Extracts the level from a log line"""
    matched = re.search(r"-\s(ERROR|WARNING|INFO|DEBUG)\s-", line)
    if matched:
        [level] = matched.groups()
        return level

    return None


def parse_logs(logs: List[str]) -> List[Dict[str, str]]:
    """Convert log lines into an object with log metadata"""
    result = []
    for log in logs:
        date = match_date(log)
        level = match_level(log)
        result.append({"date": date, "level": level, "log": log})

    return result


class LimitedSlidingBuffer(StringIO):
    def __init__(self, buffer=None, max_size=5 * 1000 * 1000) -> None:
        """A string buffer that accepts a maximum length."""
        StringIO.__init__(self, buffer)
        self.current_size = 0
        self.max_size = max_size

    def write(self, str: str):
        self.current_size += len(str)
        ret = StringIO.write(self, str)

        if self.current_size > self.max_size:
            new_size = min(int(self.current_size / 2), self.max_size)
            keep_str = self.getvalue()
            if len(keep_str) > 1:
                keep_str = keep_str[len(keep_str) - new_size :]
                self.truncate(0)
                self.seek(0)
                StringIO.write(self, keep_str)
            self.current_size = new_size

        return ret



"""
Base Jaseci object class

Naming conventions
* '_ids', '_id' - appended to lists of ids as feilds of jaseci objects
* 'obj' - used to represent a jaseci object
* 'item' - Items are kv pairs derived from item class, dont use elsewhere
* 'store' - refers to persistent store for hooks from engine, (dont use db)
"""

import copy
import json
import uuid
from datetime import datetime

from jaseci.prim.obj_mixins import Hookable
from jaseci.jsorc.memory import MemoryHook
from jaseci.utils.id_list import IdList
from jaseci.utils.json_handler import JaseciJsonEncoder, json_str_to_jsci_dict
from jaseci.utils.utils import log_var_out, logger, camel_to_snake

__version__ = "1.0.0"
element_fields = None


class Element(Hookable):
    """
    Base class for Jaseci for standard info shared across all objects types in
    Jaseci. This class also includes a method for dumping the non-general items
    as a single dictionary. The :meth:`jsci_payload` function automatically
    finds relevant non general fields and returns a dictionary of them.

    Keep in mind that when a valid hook is passed, the item writes itself on
    creation.

    Auto save should be faulse when loading from database so a new UUID is
    not randomly created and saved back to db on load

    TODO: Make below content a documentation variable and link in a few places
    Keep in mind the postfix of '_ids' on feilds of derived classes of element
    is a Jaseci convention for lists of uuids that are stored as hex and loaded
    as type UUID.

    NOTE: All fields of elements and it's derived classes are assumed to be
    json serializable types
    """

    def __init__(
        self,
        name="basic",
        kind="generic",
        auto_save=True,
        **kwargs,
    ):
        self.name = name
        self.kind = kind
        self.jid = uuid.uuid4().urn
        self.j_timestamp = datetime.utcnow().isoformat()
        self.j_type = camel_to_snake(type(self).__name__)
        Hookable.__init__(self, **kwargs)
        if self.is_master():
            self.set_master(self.jid)
        if auto_save:
            self.save()

    @property
    def id(self):
        return uuid.UUID(self.jid)

    @id.setter
    def id(self, obj):
        self.jid = obj.urn

    @property
    def timestamp(self):
        return datetime.fromisoformat(self.j_timestamp)

    @timestamp.setter
    def timestamp(self, obj):
        self.j_timestamp = obj.isoformat()

    def is_master(self, super_check=False, silent=True):
        """Check if self is a master"""
        ret = True
        if "master" not in self.j_type:
            ret = False
            if not silent:
                logger.error(f"{self} is not master")
        if super_check and "super" not in self.j_type:
            ret = False
            if not silent:
                logger.error(f"{self} does not have super master status")
        return ret

    def get_master(self):
        """Get the master of an element"""
        return self._h.get_obj(self._m_id, self.j_master)

    def duplicate(self, persist_dup: bool = False):
        """
        Duplicates elements by creating copy with new id
        Hook override to duplicate into mem / another store
        """
        kwargs = {"m_id": self._m_id, "h": self._h, "persist": persist_dup}
        dup = type(self)(**kwargs)
        id_save = dup.id
        for i in dup.__dict__.keys():
            if type(dup.__dict__[i]) == IdList:
                setattr(dup, i, IdList(parent_obj=dup, in_list=self.__dict__[i]))
            else:
                setattr(dup, i, self.__dict__[i])
        dup.id = id_save
        dup.timestamp = datetime.utcnow()
        dup.save()
        return dup

    def is_equivalent(self, obj):
        """
        Duplicates elements by creating copy with new id
        """
        if self.j_type != obj.j_type:
            return False
        for i in vars(self).keys():
            if not i.startswith("_") and not callable(getattr(self, i)):
                if i != "jid" and i != "j_timestamp":
                    if getattr(self, i) != getattr(obj, i):
                        return False
        return True

    def jsci_payload(self):
        """
        Returns all data fields and values of jaseci object as json string.
        This grabs any fields that are added into inherited objects. Useful for
        saving and loading item.
        """
        global element_fields
        if element_fields is None:
            element_fields = dir(Element(m_id=0, h=MemoryHook()))
        obj_fields = []
        for i in vars(self).keys():
            if not i.startswith("_") and i not in element_fields:
                obj_fields.append(i)
        obj_dict = {}
        for i in obj_fields:
            obj_dict[i] = getattr(self, i)
        return json.dumps(obj_dict, cls=JaseciJsonEncoder)

    def serialize(self, deep=0, detailed=False):
        """
        Serialize Jaseci object
        """
        jdict = {}
        key_fields = [
            "name",
            "kind",
            "jid",
            "j_type",
            "context",
            "code_sig",
            "j_timestamp",
            "version",
            "to_node_id",
            "from_node_id",
        ]
        for i in vars(self).keys():
            if not i.startswith("_"):
                if not detailed and i not in key_fields:
                    continue
                jdict[i] = copy.copy(vars(self)[i])
                if not detailed and i == "context":
                    for j in self.private_values():
                        del jdict[i][j]
                if deep > 0 and isinstance(jdict[i], IdList):
                    for j in range(len(jdict[i])):
                        jdict[i][j] = copy.copy(
                            self._h.get_obj(self._m_id, jdict[i][j]).serialize(deep - 1)
                        )
        return jdict

    def json(self, deep=0, detailed=False):
        """
        Returns entire self object as Json string

        deep indicates number of levels to unwind uuids
        """
        return json.dumps(
            self.serialize(deep, detailed=detailed), indent=4, cls=JaseciJsonEncoder
        )

    def json_load(self, blob):
        """Loads self from json blob"""
        jdict = json_str_to_jsci_dict(blob, parent_obj=self)
        self.dict_load(jdict=jdict)

    def dict_load(self, jdict):
        """Loads self from dict"""
        for i in jdict.keys():
            if i in vars(self).keys():
                setattr(self, i, jdict[i])

    def get_deep_obj_list(self, objs=None):
        """Recursively get all contained Jaseci objects and return id_list"""
        if objs is None:
            objs = []
        objs.append(self)
        for i in self.__dict__.keys():
            if str(i).endswith("_ids") and isinstance(self.__dict__[i], IdList):
                for j in self.__dict__[i].obj_list():
                    if j not in objs:
                        j.get_deep_obj_list(objs=objs)
        return objs

    def propagate_access(self, set_access=None):
        """Propagate permission through id_lists hierarchy"""
        if set_access is None:
            set_access = self.j_access
        for i in self.get_deep_obj_list():
            i.j_access = set_access

    def log_output(self, val):
        """
        Utility to log out variable in pretty format
        """
        log_var_out(val)

    def __str__(self):
        """
        String representation is of the form type:name:kind
        """
        return self.j_type + ":" + self.kind + ":" + self.name + ":" + self.jid

    def __repr__(self):
        return self.__str__()



"""
Module that manage and optimizes the actions configuration of Jaseci
"""
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.kube_svc import KubeService
from jaseci.jsorc.remote_actions import ACTIONS_SPEC_LOC
from jaseci.utils.utils import logger
from jaseci.jsorc.live_actions import (
    load_module_actions,
    unload_module,
    unload_remote_actions,
    load_remote_actions,
    live_actions,
    action_configs,
)

import requests
import copy
import time

from .actions_state import ActionsState

POLICIES = ["Default", "Evaluation"]


class ActionsOptimizer:
    def __init__(
        self,
        namespace: str = "default",
        policy: str = "Default",
        benchmark: dict = {},
        actions_history: dict = {},
        actions_calls: dict = {},
    ) -> None:
        self.actions_state = ActionsState()
        self.actions_change = {}
        self.jsorc_interval = 0
        self.namespace = namespace
        self.policy = policy
        self.benchmark = benchmark
        self.actions_history = actions_history
        self.actions_calls = actions_calls
        self.policy_params = {}
        self.policy_state = {}

    def kube_create(self, config):
        kube = JsOrc.svc("kube").poke(cast=KubeService)
        for kind, conf in config.items():
            name = conf["metadata"]["name"]
            kube.create(kind, name, conf, kube.namespace, "ActionsOptimzer:")

    def kube_delete(self, config):
        kube = JsOrc.svc("kube").poke(cast=KubeService)
        for kind, conf in config.items():
            name = conf["metadata"]["name"]
            kube.delete(kind, name, kube.namespace, "ActionsOptimzer:")

    def get_actions_status(self, name=""):
        """
        Return the state of action
        """
        if name == "":
            return self.actions_state.get_all_state()
        else:
            return self.actions_state.get_state(name)

    def retire_remote(self, name):
        """
        Retire a microservice through the kube service
        """
        config = action_configs[name]["remote"]
        self.kube_delete(config)
        self.actions_state.remove_remote(name)

    def spawn_remote(self, name):
        """
        Spawn a microservice through the kube service
        """
        config = action_configs[name]["remote"]
        self.kube_create(config)
        url = f"http://{config['Service']['metadata']['name']}/"
        return url

    def call_action(self, action_name, *params):
        """
        Call an action via live_actions
        """
        func = live_actions[action_name]
        func(*params)

    def action_prep(self, name):
        """
        Any action preparation that needs to be called right after action is loaded
        """
        pass

    def load_action_remote(self, name, unload_existing=False):
        """
        Load a remote action.
        JSORC will get the URL of the remote microservice and stand up a microservice if there isn't currently one in the cluster.
        Return True if the remote action is loaded successfully,
        False otherwise
        """
        cur_state = self.actions_state.get_state(name)
        if cur_state is None:
            cur_state = self.actions_state.init_state(name)

        if cur_state["mode"] == "remote" and cur_state["remote"]["status"] == "READY":
            # Check if there is already a remote action loaded
            return True

        url = self.actions_state.get_remote_url(name)
        if url is None:
            # Spawn a remote microservice
            url = self.spawn_remote(name)
            self.actions_state.start_remote_service(name, url)
            cur_state = self.actions_state.get_state(name)

        if cur_state["remote"]["status"] == "STARTING":
            if_ready = self.remote_action_ready_check(name, url)
            if if_ready:
                self.actions_state.set_remote_action_ready(name)
                cur_state = self.actions_state.get_state(name)

        if cur_state["remote"]["status"] == "READY":
            if unload_existing:
                self.unload_action_module(name)
            load_remote_actions(url)
            self.action_prep(name)
            self.actions_state.remote_action_loaded(name)
            return True

        return False

    def load_action_module(self, name, unload_existing=False):
        """
        Load an action module
        """
        cur_state = self.actions_state.get_state(name)
        if cur_state is None:
            cur_state = self.actions_state.init_state(name)

        if cur_state["mode"] == "module":
            # Check if there is already a local action loaded
            return

        if name not in action_configs:
            return

        module = action_configs[name]["module"]
        loaded_module = action_configs[name]["loaded_module"]
        if unload_existing:
            self.unload_action_remote(name)

        load_module_actions(module, loaded_module)
        self.action_prep(name)
        self.actions_state.module_action_loaded(name, module, loaded_module)

    def unload_action_auto(self, name):
        """
        Unload an action based on how it is currently loaded
        """
        cur_state = self.actions_state.get_state(name)
        if cur_state is None:
            return False, "Action is not loaded."
        if cur_state["mode"] == "module":
            return self.unload_action_module(name)
        elif cur_state["mode"] == "remote":
            return self.unload_action_remote(name)
        return False, f"Unrecognized action loaded status {cur_state['mode']}"

    def unload_action_module(self, name):
        """
        Unload an action module
        """
        cur_state = self.actions_state.get_state(name)
        if cur_state is None:
            return False, "Action is not loaded."

        if cur_state["mode"] != "module":
            return False, "Action is not loaded as module."

        module_name = cur_state["module"]["name"]
        loaded_module = cur_state["module"]["loaded_module"]

        unload_module(module_name)
        unload_module(loaded_module)
        self.actions_state.module_action_unloaded(name)

        return (True, f"Action module {name} unloaded.")

    def unload_action_remote(self, name):
        """
        Unload a remote action
        """
        cur_state = self.actions_state.get_state(name)
        if cur_state is None:
            return False, "Action is not loaded."

        if cur_state["mode"] != "remote":
            return False, "Action is not loaded as remote."

        if cur_state["remote"]["status"] != "READY":
            return False, "Remote action is not ready."

        # Get the list of actions from the action spec of the server
        url = cur_state["remote"]["url"]

        unload_remote_actions(url)
        self.actions_state.remote_action_unloaded(name)

        return (True, f"Remote actions from {url} unloaded.")

    def remote_action_ready_check(self, name, url):
        """
        Check if a remote action is ready by querying the action_spec endpoint
        """
        if url is None:
            return False
        spec_url = url.rstrip("/") + ACTIONS_SPEC_LOC
        headers = {"content-type": "application/json"}
        try:
            res = requests.get(spec_url, headers=headers, timeout=1)
        except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
            # Remote service not ready yet
            return False

        return res.status_code == 200

    def set_action_policy(self, policy_name: str, policy_params: dict = {}):
        """
        Set the action optimization policy for JSORC
        """
        # TODO: manage policy switching if there are unresolved actions state
        if policy_name in POLICIES:
            self.policy = policy_name
            self.policy_state[policy_name] = {}
            self.policy_params = policy_params
            return True
        else:
            return f"Policy {policy_name} not found."

    def get_action_policy(self):
        """
        Return the currently active action policy
        """
        return self.policy

    def run(self, jsorc_interval: int):
        """
        The main optimization function.
        This gets invoked by JSROC regularly at a configured interval.
        """
        self.jsorc_interval = jsorc_interval
        if self.policy == "Default":
            # Default policy does not manage action automatically
            return
        elif self.policy == "Evaluation":
            self._actionpolicy_evaluation()

        if len(self.actions_change) > 0:
            self.apply_actions_change()

    def _init_evalution_policy(self, policy_state):
        # 999 is just really large memory size so everything can fits in local
        node_mem = self.policy_params.get("node_mem", 999 * 1024)
        jaseci_runtime_mem = self.policy_params.get("jaseci_runtime_mem", 300)
        # Initialize configs to eval
        actions = self.actions_state.get_active_actions()
        # construct list of possible configurations
        all_configs = [{"local_mem": jaseci_runtime_mem}]
        for act in actions:
            new_configs = []
            for con in all_configs:
                for m in ["local", "remote"]:
                    c = copy.deepcopy(con)
                    c[act] = m
                    if m == "local":
                        c["local_mem"] += action_configs[act].get("remote_memory", 0)
                        if c["local_mem"] < node_mem:
                            new_configs.append(dict(c))
                    else:
                        new_configs.append(dict(c))
            all_configs = list(new_configs)
        policy_state["remain_configs"] = all_configs

    def _actionpolicy_evaluation(self):
        """
        A evaluation based policy.
        JSORC cycle through possible action configurations and evaluate request performance and select the one with the best performance.
        Use the post_request_hook from JSORC to track request performance
        """
        logger.info("===Evaluation Policy===")
        policy_state = self.policy_state["Evaluation"]

        if len(policy_state) == 0:
            # Initialize policy tracking state
            policy_state = {
                "phase": "eval",  # current phase of policy: eval|perf
                "cur_config": None,  # current active configuration
                "remain_configs": [],  # remaining configurations that need to be evaluated
                "past_configs": [],  # configurations already evaluated
                "eval_phase": 10,  # how long is evaluatin period (in seconds)
                "perf_phase": 100,  # how long is the performance period (in seconds)
                "cur_phase": 0,  # how long the current period has been running
                "prev_best_config": self.actions_state.get_all_state(),
            }
        policy_state["cur_phase"] += self.jsorc_interval

        # check if we should go into evaluation phase
        if (
            policy_state["phase"] == "perf"
            and policy_state["cur_phase"] >= policy_state["perf_phase"]
        ):
            # if no enough walker were execueted in this period, keep in perf phase
            if "walker_run" not in self.benchmark["requests"]:
                policy_state["cur_phase"] = 0
            else:
                logger.info("===Evaluation Policy=== Switching to evaluation mode")
                policy_state["phase"] = "eval"
                policy_state["cur_phase"] = 0
                policy_state["cur_config"] = None
                if len(policy_state["remain_configs"]) == 0:
                    self._init_evalution_policy(policy_state)
        if policy_state["phase"] == "eval":
            # In evaluation phase
            if policy_state["cur_config"] is None:
                self._init_evalution_policy(policy_state)

                # This is the start of evaluation period
                policy_state["cur_config"] = policy_state["remain_configs"][0]
                del policy_state["remain_configs"][0]
                policy_state["cur_phase"] = 0
                self.benchmark["active"] = True
                self.benchmark["requests"] = {}
                self.actions_change = self._get_action_change(
                    policy_state["cur_config"]
                )
                if len(self.actions_change) > 0:
                    logger.info(
                        f"===Evaluation Policy=== Switching eval config to {policy_state['cur_config']}"
                    )
                    policy_state["phase"] = "eval_switching"
                    self.benchmark["active"] = False
            else:
                if policy_state["cur_phase"] >= policy_state["eval_phase"]:
                    # The eval phase for the current configuration is complete
                    # Get performance
                    if "walker_run" not in self.benchmark["requests"]:
                        # meaning no incoming requests during this period.
                        # stay in this phase
                        logger.info(f"===Evaluation Policy=== No walkers were executed")
                        self.policy_state["Evaluation"] = policy_state
                        return

                    walker_runs = []
                    for walker, times in self.benchmark["requests"][
                        "walker_run"
                    ].items():
                        if walker == "_default_":
                            continue
                        else:
                            walker_runs.extend(times)

                    avg_walker_lat = sum(walker_runs) / len(walker_runs)
                    policy_state["cur_config"]["avg_walker_lat"] = avg_walker_lat
                    policy_state["past_configs"].append(policy_state["cur_config"])
                    logger.info(
                        f"===Evaluation Policy=== Complete evaluation period for {policy_state['cur_config']} latency: {avg_walker_lat}"
                    )

                    # check if all configs have been evaluated
                    if len(policy_state["remain_configs"]) == 0:
                        # best config is the one with the fastest walker latency during the evaluation period
                        logger.info(f"===Evaluation Policy=== Evaluation phase over. ")
                        best_config = min(
                            policy_state["past_configs"],
                            key=lambda x: x["avg_walker_lat"],
                        )
                        # Switch the system to the best config
                        del best_config["avg_walker_lat"]
                        self.actions_change = self._get_action_change(best_config)

                        # ADAPTIVE: if the selected best config is the same config as the previous best one, double the performance period
                        if all(
                            [
                                best_config[act]
                                == policy_state["prev_best_config"][act]["mode"]
                                for act in best_config.keys()
                                if act in action_configs.keys()
                            ]
                        ):
                            policy_state["perf_phase"] *= 2
                            logger.info(
                                f"===Evaluation Policy=== Best config is the same as previous one. Doubling performance phase to {policy_state['perf_phase']}"
                            )

                        policy_state["phase"] = "perf"
                        policy_state["cur_config"] = None
                        policy_state["past_configs"] = []
                        policy_state["cur_phase"] = 0
                        self.benchmark["requests"] = {}
                        self.benchmark["active"] = True
                        logger.info(
                            f"===Evaluation Policy=== Evaluation phase over. Selected best config as {best_config}"
                        )
                    else:
                        next_config = policy_state["remain_configs"][0]
                        del policy_state["remain_configs"][0]
                        self.actions_change = self._get_action_change(next_config)
                        policy_state["cur_config"] = next_config
                        policy_state["cur_phase"] = 0
                        self.benchmark["requests"] = {}
                        if len(self.actions_change) > 0:
                            logger.info(
                                f"===Evaluation Policy=== Switching eval config to {policy_state['cur_config']}"
                            )
                            policy_state["phase"] = "eval_switching"
                            self.benchmark["active"] = False
                        else:
                            policy_state["phase"] = "eval"
                            self.benchmark["active"] = True
                        logger.info(
                            f"===Evaluation Policy=== Switching to next config to evaluate {next_config}"
                        )
        elif policy_state["phase"] == "eval_switching":
            # in the middle of switching between configs for evaluation
            if len(self.actions_change) == 0:
                # this means all actions change have been applied, start evaluation phase
                logger.info(
                    f"===Evaluation Policy=== All actions change have been applied. Start evaluation phase."
                )
                policy_state["phase"] = "eval"
                policy_state["cur_phase"] = 0
                self.benchmark["active"] = True
                self.benchmark["requests"] = {}

        self.policy_state["Evaluation"] = policy_state

    def _get_action_change(self, new_action_state):
        """
        Given a new desired action state and the current action_state tracking, return the change set
        """
        change_state = {}
        for name, new_state in new_action_state.items():
            if name not in action_configs.keys():
                continue
            cur = self.actions_state.get_state(name)
            if cur is None:
                cur = self.actions_state.init_state(name)
            if new_state == "local":
                new_state = "module"
            if new_state != cur["mode"]:
                change_str = (
                    f"{cur['mode'] if cur['mode'] is not None else ''}_to_{new_state}"
                )
                change_state[name] = change_str
        return change_state

    def apply_actions_change(self):
        """
        Apply any action configuration changes
        """
        actions_change = dict(self.actions_change)
        # For now, to_* and *_to_* are the same logic
        # But this might change down the line
        for name, change_type in actions_change.items():
            logger.info(f"==Actions Optimizer== Changing {name} {change_type}")
            if change_type == "to_local" or change_type == "to_module":
                # Switching from no action loaded to local
                self.load_action_module(name)
                del self.actions_change[name]
            elif change_type == "to_remote":
                loaded = self.load_action_remote(name)
                if loaded:
                    del self.actions_change[name]
            elif change_type == "local_to_remote" or change_type == "module_to_remote":
                # loaded = self.load_action_remote(name, unload_existing=True)
                loaded = self.load_action_remote(name)
                if loaded:
                    del self.actions_change[name]
            elif change_type == "remote_to_local" or change_type == "remote_to_module":
                # self.load_action_module(name, unload_existing=True)
                self.load_action_module(name)
                del self.actions_change[name]

        if len(actions_change) > 0 and self.actions_history["active"]:
            # Summarize action stats during this period and add to previous state
            self.summarize_action_calls()
            self.actions_history["history"].append(
                {"ts": time.time(), "actions_state": self.actions_state.get_all_state()}
            )

    def summarize_action_calls(self):
        actions_summary = {}
        for action_name, calls in self.actions_calls.items():
            actions_summary[action_name] = sum(calls) / len(calls)
        self.actions_calls.clear()

        if len(self.actions_history["history"]) > 0:
            self.actions_history["history"][-1]["actions_calls"] = actions_summary



from jaseci.utils.log_utils import LimitedSlidingBuffer, parse_logs
from jaseci.utils.test_core import CoreTest


class LoggerApiTest(CoreTest):
    fixture_src = __file__

    logs = [
        "2022-11-26 21:43:55,354 - ERROR - start: Skipping MailService due to initialization failure!",
        "ConnectionRefusedError: [Errno 61] Connection refused",
    ]

    def test_parse_logs(self):
        parsed_logs = parse_logs(self.logs)

        self.assertEqual(len(parsed_logs), 2)

        self.assertEqual(parsed_logs[0]["level"], "ERROR")
        self.assertEqual(parsed_logs[1]["level"], None)

        self.assertEqual(parsed_logs[0]["date"], "2022-11-26 21:43:55")
        self.assertEqual(parsed_logs[1]["date"], None)

    def test_get_logs(self):
        self.logger_on()
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("test_logging.jac")}],
        )
        self.logger_off()

        ret = self.call(self.smast, ["logger_get", {}])
        self.assertGreaterEqual(len(ret), 1)
        self.assertTrue(ret[-1]["log"].endswith("Hello world!"))

    def test_limited_sliding_buffer(self):
        buffer = LimitedSlidingBuffer(max_size=25)
        buffer.write("Hello world!")
        buffer.write("Hi there!")
        buffer.write("Hello world!")
        buffer.write("Hi there!")
        self.assertEqual(buffer.getvalue(), "ere!Hello world!Hi there!")
        self.assertEqual(buffer.current_size, 25)



from jaseci.jac.ir.passes.ir_pass import IrPass  # noqa
from jaseci.jac.ir.passes.pt_prune_pass import ParseTreePrunePass  # noqa
from jaseci.jac.ir.passes.printer_pass import PrinterPass  # noqa
from jaseci.jac.ir.passes.stats_pass import StatsPass  # noqa
from jaseci.jac.ir.passes.codegen_pass import CodeGenPass  # noqa
from jaseci.jac.ir.passes.ast_prune_pass import AstPrunePass  # noqa



from jaseci.utils.test_core import CoreTest, jac_testcase


class UrlTest(CoreTest):
    fixture_src = __file__

    @jac_testcase("url.jac", "is_valid_test")
    def test_is_valid(self, ret):
        self.assertEqual(
            ret["report"],
            [True, False, False],
        )

    @jac_testcase("url.jac", "ping_test")
    def test_ping(self, ret):
        self.assertEqual(
            ret["report"],
            [True, False, False, False],
        )

    @jac_testcase("url.jac", "download_text_test")
    def test_download_text(self, ret):
        # Download the jaseci homepage as a test
        self.assertGreater(abs(len(ret["report"][0])), 0)

    @jac_testcase("url.jac", "download_b64_test")
    def test_download_b64(self, ret):
        # Jaseci white paper is 2,034,680 bytes in base64
        self.assertEqual(
            len(ret["report"][0]),
            2034680,
        )



"""
Architype class for Jaseci

Each architype is a registered templatized version of instances of any Jaseci
abstractions or collections of instances (e.g., subgraphs, etc)
"""
from jaseci.prim.element import Element
from jaseci.jac.interpreter.architype_interp import ArchitypeInterp
from jaseci.jac.ir.jac_code import JacCode
from jaseci.utils.id_list import IdList


class Architype(Element, JacCode, ArchitypeInterp):
    """Architype class for Jaseci"""

    def __init__(self, code_ir=None, is_async=False, *args, **kwargs):
        self.super_archs = list()
        self.anchor_var = None
        self.private_vars = []
        self.has_vars = []
        self.entry_action_ids = IdList(self)
        self.activity_action_ids = IdList(self)
        self.exit_action_ids = IdList(self)

        # async handling for walker
        self.is_async = is_async

        Element.__init__(self, *args, **kwargs)
        JacCode.__init__(self, code_ir)
        ArchitypeInterp.__init__(self)

    def run(self):
        """
        Create set of new object instances from architype if needed
        """
        return self.run_architype(jac_ast=self.get_jac_ast())

    def get_all_actions(self):
        actions = IdList(self, auto_save=False)
        for i in self.arch_with_supers():
            actions += i.entry_action_ids + i.activity_action_ids + i.exit_action_ids
        return actions

    def arch_with_supers(self):
        archs = [self]
        for i in self.super_archs:
            obj = self.parent().arch_ids.get_obj_by_name(name=i, kind=self.kind)
            archs += obj.arch_with_supers()
        return archs

    def derived_types(self):
        names = []
        for i in self.arch_with_supers():
            names += i.super_archs + [i.name]
        return names

    def is_instance(self, name):
        return name in self.derived_types()

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        des = (
            self.activity_action_ids.obj_list()
            + self.entry_action_ids.obj_list()
            + self.exit_action_ids.obj_list()
        )
        for i in des:
            i.destroy()
        super().destroy()



"""Built in actions for Jaseci"""
import requests
from jaseci.jsorc.live_actions import jaseci_action
from bs4 import BeautifulSoup


@jaseci_action()
def get_page_meta(url: str):
    """
    Util to parse metadata out of urls and html documents
    """
    try:
        webpage = requests.get(url)
        soup = BeautifulSoup(webpage.content, features="lxml")
        meta = soup.find_all("meta")
        meta_list = []
        for tag in meta:
            meta_list.append(dict(tag.attrs))
        return meta_list
    except Exception as e:
        print("Failed")
        return f"Failed at getting metadata for {url}: {str(e)}"



"""
Prometheus APIs
"""
from jaseci.extens.api.interface import Interface
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.prome_svc import PrometheusService


def prome():
    return JsOrc.svc("prome").poke(PrometheusService)


class PrometheusApi:
    """
    Prometheus APIs
    """

    @Interface.admin_api()
    def prometheus_metrics_list(self):
        """
        Return list of availabel metrics
        """
        return prome().all_metrics()

    @Interface.admin_api()
    def prometheus_pod_list(self, namespace: str = "", exclude_prom: bool = False):
        """
        Return list of pods. If exclude_prom,
        """
        return prome().pods(namespace=namespace, exclude_prom=exclude_prom)

    @Interface.admin_api()
    def prometheus_pod_info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ):
        """
        Return pods info and metrics
        """
        return (
            prome()
            .info(
                namespace=namespace,
                exclude_prom=exclude_prom,
                timestamp=timestamp,
                duration=duration,
            )
            .pod
        )

    @Interface.admin_api()
    def prometheus_node_info(
        self,
        namespace: str = "",
        exclude_prom: bool = False,
        timestamp: int = 0,
        duration: int = 0,
    ):
        """
        Return pods info and metrics
        """
        return (
            prome()
            .info(
                namespace=namespace,
                exclude_prom=exclude_prom,
                timestamp=timestamp,
                duration=duration,
            )
            .node
        )



from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.test_core import CoreTest
from unittest.mock import patch, Mock
from kubernetes import config as kubernetes_config, client
from jaseci.jsorc.jsorc_settings import JsOrcSettings
from .fixture.elastic_resources import ELASTIC_RESOURCES
from urllib3.response import HTTPResponse


def aggregate_k8s_resources(mocked):
    """Aggregate k8s resources from multiple mocked calls to kube"""
    resources = {}
    call_args_list = mocked.call_args_list
    for call_args in call_args_list:
        args, _ = call_args
        # arg1: kind, arg2: name
        kind = args[0]
        name = args[1]
        if kind not in resources:
            resources[kind] = []
        # The resource name for Namespace is different from run to run
        if kind != "Namespace":
            resources[kind].append(name)

    return resources


class MockKubeTest(CoreTest):
    """Custom test class that mock the environment to be inside a k8s cluster"""

    @patch.object(client.ApiClient, "call_api")
    @patch.object(kubernetes_config, "load_incluster_config")
    def setUp(self, mock_config, mock_call_api):
        super().setUp()
        JsOrcSettings.KUBE_CONFIG["enabled"] = True
        JsOrcSettings.ELASTIC_CONFIG["enabled"] = True
        self.kube = JsOrc.svc("kube")


class JsOrcTest(MockKubeTest):
    """Unit test for JsOrc"""

    def setUp(self):
        super().setUp()

    def test_jsorc_elastic_create(self):
        self.call(
            self.smast,
            [
                "config_update",
                {"name": "ELASTIC_CONFIG", "field_key": "enabled", "field_value": True},
            ],
        )
        with patch.object(self.kube, "read") as mock_read, patch.object(
            self.kube, "patch"
        ) as mock_patch, patch.object(self.kube, "create") as mock_create:
            # resource do not exist
            mocked_response = HTTPResponse(status=404)
            mock_read.return_value = mocked_response
            mock_patch.return_value = Mock()
            mock_create.return_value = Mock()

            JsOrc.svc_reset("elastic")

            # Mocking elastic health check to show its up and running
            with patch("jaseci.extens.svc.elastic_svc.Elastic._get") as mocked_get:
                mocked_get.return_value = {"timed_out": False}
                JsOrc.regenerate()

            read_resources = aggregate_k8s_resources(mock_read)
            assert read_resources == ELASTIC_RESOURCES

            create_resources = aggregate_k8s_resources(mock_create)
            assert create_resources == ELASTIC_RESOURCES

            # No resources should be patched
            assert not mock_patch.called

    def test_jsorc_elastic_patch(self):
        self.call(
            self.smast,
            [
                "config_update",
                {"name": "ELASTIC_CONFIG", "field_key": "enabled", "field_value": True},
            ],
        )
        with patch.object(self.kube, "read") as mock_read, patch.object(
            self.kube, "patch"
        ) as mock_patch, patch.object(self.kube, "create") as mock_create:
            mock_read.return_value = Mock()
            mock_patch.return_value = Mock()
            mock_create.return_value = Mock()

            JsOrc.svc_reset("elastic")

            # Mocking elastic health check to show its up and running
            with patch("jaseci.extens.svc.elastic_svc.Elastic._get") as mocked_get:
                mocked_get.return_value = {"timed_out": False}
                JsOrc.regenerate()

            read_resources = aggregate_k8s_resources(mock_read)
            assert read_resources == ELASTIC_RESOURCES

            patch_resources = aggregate_k8s_resources(mock_patch)
            # No need to patch Namespace
            del ELASTIC_RESOURCES["Namespace"]

            assert patch_resources == ELASTIC_RESOURCES

            # No resources should be created
            assert not mock_create.called



"""
Graph  class for Jaseci

"""
from jaseci.prim.node import Node
from jaseci.utils.id_list import IdList


class Graph(Node):
    """Graph class for Jaseci"""

    def __init__(self, **kwargs):
        self.hd_node_ids = IdList(self)
        Node.__init__(self, **kwargs)
        self.name = "root"
        self.kind = "node"

        # Node.__init__(self, name="root", kind="node", **kwargs)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.hd_node_ids.obj_list():
            i.destroy()
        super().destroy()



"""
General action base class with automation for hot loading
"""
from importlib.util import spec_from_file_location, module_from_spec
from jaseci.utils.utils import logger
from jaseci.jsorc.remote_actions import ACTIONS_SPEC_LOC
from jaseci.jsorc.remote_actions import serv_actions, mark_as_remote, mark_as_endpoint
import requests
import os
import sys
import inspect
import importlib
import gc

live_actions = {}  # {"act.func": func_obj, ...}
live_action_modules = {}  # {__module__: ["act.func1", "act.func2", ...], ...}
action_configs = {}  # {"module_name": {}, ...}


def jaseci_action(act_group=None, aliases=list(), allow_remote=False):
    """Decorator for Jaseci Action interface"""
    caller_globals = dict(inspect.getmembers(inspect.currentframe().f_back))[
        "f_globals"
    ]
    if allow_remote and "serv_actions" not in caller_globals:
        caller_globals["serv_actions"] = serv_actions

    def decorator_func(func):
        if allow_remote:
            mark_as_remote([func, act_group, aliases, caller_globals])
        return assimilate_action(func, act_group, aliases)

    return decorator_func


def jaseci_expose(endpoint, mount=None):
    """Decorator for Jaseci Action interface"""
    caller_globals = dict(inspect.getmembers(inspect.currentframe().f_back))[
        "f_globals"
    ]
    if "serv_actions" not in caller_globals:
        caller_globals["serv_actions"] = serv_actions

    def decorator_func(func):
        mark_as_endpoint([func, endpoint, mount, caller_globals])
        return func

    return decorator_func


def assimilate_action(func, act_group=None, aliases=list()):
    """Helper for jaseci_action decorator"""
    act_group = [func.__module__.split(".")[-1]] if act_group is None else act_group
    action_name = f"{'.'.join(act_group+[func.__name__])}"
    live_actions[action_name] = func
    if func.__module__ != "js_remote_hook":
        if func.__module__ in live_action_modules:
            live_action_modules[func.__module__].append(action_name)
        else:
            live_action_modules[func.__module__] = [action_name]
    for i in aliases:
        live_actions[f"{'.'.join(act_group+[i])}"] = func
        if func.__module__ != "js_remote_hook":
            if func.__module__ in live_action_modules:
                live_action_modules[func.__module__].append(
                    f"{'.'.join(act_group+[i])}"
                )
            else:
                live_action_modules[func.__module__] = [f"{'.'.join(act_group+[i])}"]
    return func


def load_local_actions(file: str, ctx: dict = {}):
    """Load all jaseci actions from python file"""
    try:
        name = file.rstrip(".py")
        name = ".".join(name.split("/")[-2:])
        # Assumes parent folder of py file is a package for internal relative
        # imports, name is package.module and package path is added to sys path
        spec = spec_from_file_location(name, str(file))
        if spec is None:
            logger.error(f"Cannot hot load from action file {file}")
            return False
        else:
            module_dir = os.path.dirname(os.path.dirname(os.path.realpath(file)))
            if module_dir not in sys.path:
                sys.path.append(module_dir)
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)
            try:
                if hasattr(mod, "setup"):
                    mod.setup(**ctx)
            except Exception as e:
                logger.error(
                    f"Cannot run set up for module {mod}. This could be because the module doesn't have a setup procedure for initialization, or wrong setup parameters are provided."
                )
                logger.error(e)
            return True
    except Exception as e:
        logger.error(f"Cannot hot load local actions from {file}: {e}")
        return False


def load_module_actions(mod, loaded_module=None, ctx: dict = {}):
    """Load all jaseci actions from python module"""
    try:
        if mod in sys.modules:
            del sys.modules[mod]
        if loaded_module and loaded_module in sys.modules:
            del sys.modules[loaded_module]
        if mod in live_action_modules:
            for i in live_action_modules[mod]:
                if i in live_actions:
                    del live_actions[i]
        if loaded_module in live_action_modules:
            for i in live_action_modules[loaded_module]:
                if i in live_actions:
                    del live_actions[i]

        mod = importlib.import_module(mod)
        try:
            if hasattr(mod, "setup"):
                mod.setup(**ctx)
        except Exception as e:
            logger.error(
                f"Cannot run set up for module {mod}. This could be because the module doesn't have a setup procedure for initialization, or wrong setup parameters are provided."
            )
            logger.error(e)
        if mod:
            return True
    except Exception as e:
        logger.error(f"Cannot hot load module actions from {mod}: {e}")

    return False


def load_action_config(config, module_name):
    """
    Load the action config of a jaseci action module
    """

    loaded_configs = importlib.import_module(config).ACTION_CONFIGS
    if module_name and module_name in loaded_configs:
        action_configs[module_name] = loaded_configs[module_name]
        return True
    else:
        return False


def unload_module(mod):
    """Unload actions module and all relevant function"""
    if mod in sys.modules.keys() and mod in live_action_modules.keys():
        for i in live_action_modules[mod]:
            if i in live_actions:
                del live_actions[i]

        # Iterate through the objects in the module __dict__ to manually delete them
        loaded_mod = sys.modules[mod]
        mod_content_len = len(loaded_mod.__dict__)
        for _ in range(mod_content_len):
            mod_obj = loaded_mod.__dict__.pop(list(loaded_mod.__dict__.keys())[0])
            del mod_obj
        del loaded_mod
        del sys.modules[mod]
        del live_action_modules[mod]
        gc.collect()
        return True
    return False


def unload_action(name):
    """Unload actions module and all relevant function"""
    if name in live_actions.keys():
        mod = live_actions[name].__module__
        if mod != "js_remote_hook":
            live_action_modules[mod].remove(name)
            if len(live_action_modules[mod]) < 1:
                unload_module(mod)
        del live_actions[name]
        return True
    return False


def unload_actionset(name):
    """Unload actions module and all relevant function"""
    act_list = []
    orig_len = len(live_actions)
    for i in live_actions.keys():
        if i.startswith(name + "."):
            act_list.append(i)
    for i in act_list:
        unload_action(i)
    return len(live_actions) != orig_len


def load_preconfig_actions(hook):
    import json

    action_preload = hook.resolve_glob("ACTION_SETS", None)
    if action_preload:
        try:
            action_preload = json.loads(action_preload)
            for i in action_preload["local"]:
                load_local_actions(i)
            for i in action_preload["remote"]:
                load_remote_actions(i)
            for i in action_preload["module"]:
                load_module_actions(i)
        except Exception:
            pass


def get_global_actions():
    """
    Loads all global action hooks for use by Jac programs
    Attaches globals to mem_hook
    """
    from jaseci.prim.action import Action
    from jaseci.jsorc.memory import MemoryHook

    global_action_list = []
    hook = MemoryHook()
    for i in live_actions.keys():
        if (
            i.startswith("std.")
            or i.startswith("file.")
            or i.startswith("net.")
            or i.startswith("rand.")
            or i.startswith("vector.")
            or i.startswith("request.")
            or i.startswith("date.")
            or i.startswith("jaseci.")
            or i.startswith("internal.")
            or i.startswith("zip.")
            or i.startswith("webtool.")
            or i.startswith("url.")
            or i.startswith("regex.")
        ):
            global_action_list.append(
                Action(
                    m_id=0,
                    h=hook,
                    mode="public",
                    name=i,
                    value=i,
                    persist=False,
                )
            )
    return global_action_list


def unload_remote_actions(url):
    """
    Get the list of actions from the given URL and then unload them.
    """
    headers = {"content-type": "application/json"}
    try:
        spec = requests.get(url.rstrip("/") + ACTIONS_SPEC_LOC, headers=headers)
        spec = spec.json()
        for i in spec.keys():
            unload_action(i)
        return True
    except Exception as e:
        logger.error(f"Cannot unload remote action from {url}: {e}")


def load_remote_actions(url, ctx: dict = {}):
    """Load all jaseci actions from live pod"""
    headers = {"content-type": "application/json"}
    try:
        spec = requests.get(url.rstrip("/") + ACTIONS_SPEC_LOC, headers=headers)
        spec = spec.json()
        for i in spec.keys():
            live_actions[i] = gen_remote_func_hook(url, i, spec[i])
            if i.endswith(".setup") and ctx:
                try:
                    live_actions[i](**ctx)
                except Exception as e:
                    logger.error(
                        f"Cannot run set up for remote action {i}. This could be because the module doesn't have a setup procedure for initialization, or wrong setup parameters are provided."
                    )
                    logger.error(e)
        return True

    except Exception as e:
        logger.error(f"Cannot hot load remote actions at {url}: {e}")
        return False


def gen_remote_func_hook(url, act_name, param_names):
    """Generater for function calls for remote action calls"""

    def func(*args, **kwargs):
        params = {}
        for i in range(len(param_names)):
            if i < len(args):
                params[param_names[i]] = args[i]
            else:
                params[param_names[i]] = None
        for i in kwargs.keys():
            if i in param_names:
                params[i] = kwargs[i]
        # Remove any None-valued parameters to use the default value of the action def
        params = dict([(k, v) for k, v in params.items() if v is not None])
        act_url = f"{url.rstrip('/')}/{act_name.split('.')[-1]}"
        res = requests.post(
            act_url, headers={"content-type": "application/json"}, json=params
        )
        return res.json()

    func.__module__ = "js_remote_hook"

    return func


def call_action(action_name: str, ctx: dict = {}) -> None:
    """
    Call an action by name
    """
    try:
        action_name = action_name.strip()
        if action_name in live_actions.keys():
            res = live_actions[action_name](**ctx)
            return res, True
        else:
            raise Exception(f"Action {action_name} not found")
    except Exception as e:
        logger.error(f"Error calling action {action_name}: {e}")
        return None, False



"""
Variable manager for Jac

Representations for all jac runtime variables
"""
from jaseci.prim.element import Element
from jaseci.prim.obj_mixins import Anchored
from jaseci.prim.node import Node
from jaseci.prim.edge import Edge
from jaseci.prim.graph import Graph
from jaseci.jac.jac_set import JacSet
import uuid

NoneType = type(None)


class JacType:
    STR = "JAC_TYPE.STR"
    INT = "JAC_TYPE.INT"
    FLOAT = "JAC_TYPE.FLOAT"
    LIST = "JAC_TYPE.LIST"
    DICT = "JAC_TYPE.DICT"
    BOOL = "JAC_TYPE.BOOL"
    NODE = "JAC_TYPE.NODE"
    EDGE = "JAC_TYPE.EDGE"
    TYPE = "JAC_TYPE.TYPE"
    NULL = "JAC_TYPE.NULL"


def jac_type_wrap(val):
    if isinstance(val, type):
        if val == str:
            val = JacType.STR
        elif val == int:
            val = JacType.INT
        elif val == float:
            val = JacType.FLOAT
        elif val == list:
            val = JacType.LIST
        elif val == JacSet:
            val = JacType.LIST
        elif val == dict:
            val = JacType.DICT
        elif val == bool:
            val = JacType.BOOL
        elif val in [Node, Graph]:
            val = JacType.NODE
        elif val == Edge:
            val = JacType.EDGE
        elif val == type:
            val = JacType.TYPE
        elif val == NoneType:
            val = JacType.NULL
    return val


def jac_type_unwrap(val):
    if type(val) == str and val.startswith("JAC_TYPE."):
        if val == JacType.STR:
            val = str
        elif val == JacType.INT:
            val = int
        elif val == JacType.FLOAT:
            val = float
        elif val == JacType.LIST:
            val = list
        elif val == JacType.DICT:
            val = dict
        elif val == JacType.BOOL:
            val = bool
        elif val == JacType.NODE:
            val = Node
        elif val == JacType.EDGE:
            val = Edge
        elif val == JacType.TYPE:
            val = type
        elif val == JacType.NULL:
            val = NoneType
    return val


def is_jac_elem(val):
    """Test if value is jac element"""
    return type(val) == str and val.startswith("jac:uuid:")


def jac_elem_wrap(val, serialize_mode=False):
    if serialize_mode:
        if isinstance(val, Anchored):
            val.context = jac_wrap_value(val.context)
        val = val.serialize()
    else:
        val = val.jid.replace("urn", "jac")
    return val


def jac_elem_unwrap(val, parent):
    val = parent._h.get_obj(parent._m_id, val.replace("jac", "urn"))
    return val


def jac_wrap_value(val, serialize_mode=False):
    """converts all elements to uuids in lists etc"""
    val = jac_type_wrap(val)
    if isinstance(val, Element):
        val = jac_elem_wrap(val, serialize_mode=serialize_mode)
    elif isinstance(val, JacSet):
        val = jac_wrap_value(list(val), serialize_mode)
    elif isinstance(val, list):
        for i in range(len(val)):
            val[i] = jac_wrap_value(val[i], serialize_mode)
    elif isinstance(val, dict):
        for i in val.keys():
            val[i] = jac_wrap_value(val[i], serialize_mode)
    return val


def jac_unwrap_value(val, parent):
    """Reference to variables value"""
    if is_jac_elem(val):
        val = jac_elem_unwrap(val, parent=parent)
    elif isinstance(val, list):
        for i in range(len(val)):
            val[i] = jac_unwrap_value(val[i], parent=parent)
    elif isinstance(val, dict):
        for i in val.keys():
            val[i] = jac_unwrap_value(val[i], parent=parent)
    return jac_type_unwrap(val)


class JacValue:
    """
    A reference to a variable in context dict that is common for elements
    """

    def __init__(
        self, parent, value=None, ctx=None, name=None, end=None, create_mode=False
    ):
        """
        Abstraction of all Jac types, ctx and name serve as obj[key/idx]
        end is for idx ranges for list slices
        """
        self.parent = parent
        self.ctx = ctx
        self.is_element = False
        self.name = name
        self.end = end
        self.value = self.setup_value(value, create_mode=create_mode)

    def setup_value(self, value, create_mode):
        if isinstance(self.ctx, Element):
            self.is_element = self.ctx
            if self.parent._assign_mode:
                self.is_element.save()
            self.ctx = self.ctx.context
        if value is not None:
            return value
        elif self.ctx is not None and self.name is not None:
            if self.end is not None:
                return self.ctx[self.name : self.end]
            elif type(self.name) == int or self.name in self.ctx.keys():
                return self.ctx[self.name]
            elif (
                not self.parent._assign_mode
                and not create_mode
                and (
                    not self.is_element
                    or self.name not in self.is_element.get_architype().has_vars
                )
            ):
                self.parent.rt_error(f"Key {self.name} not found in object/dict.")
        else:
            return None

    def write(self, jac_ast, force=False):
        if (
            not force
            and self.is_element
            and self.name not in self.ctx.keys()
            and self.name not in self.is_element.get_architype().has_vars
        ):
            self.parent.rt_error(
                f"Creating variable {self.name} in graph "
                f"element {type(self.is_element)} is not allowed, "
                "please define",
                jac_ast,
            )
        elif self.ctx is None or self.name is None:
            self.parent.rt_error(
                f"No valid live variable! ctx: {self.ctx} name: {self.name}", jac_ast
            )
        elif self.end is not None:
            self.ctx[self.name : self.end] = self.value
        else:
            self.ctx[self.name] = self.value

    def check_assignable(self, jac_ast=None):
        if self.ctx is None:
            self.parent.rt_error("Cannot assign to this value", jac_ast)
            return False
        return True

    def self_destruct(self, jac_ast):
        if self.is_element and self.name in self.ctx.keys():
            self.ctx[self.name] = None  # assumes interp has destroyed element
        elif self.ctx is not None:
            try:
                del self.ctx[self.name]
            except Exception as e:
                self.parent.rt_error(f"{e}", jac_ast)
        else:
            self.parent.rt_error(
                f"{self.value} is not destroyable, try setting to null", jac_ast
            )

    def jac_type(self):
        """Return Jac type of value"""
        return jac_type_wrap(type(self.value))

    def __str__(self):
        return (
            "JacValue:" + str(self.ctx) + ":" + str(self.name) + ":" + str(self.value)
        )



"""
Interpreter for jac code in AST form

This interpreter should be inhereted from the class that manages state
referenced through self.
"""
from copy import copy
from jaseci.utils.utils import logger
from jaseci.jsorc.live_actions import live_actions, load_preconfig_actions

# from jaseci.actions.find_action import find_action
from jaseci.prim.element import Element

from jaseci.jac.jac_set import JacSet
from jaseci.jac.machine.jac_scope import JacScope
from jaseci.utils.id_list import IdList
from jaseci.jac.ir.ast import Ast
from jaseci.prim.edge import Edge
from jaseci.prim.node import Node
from jaseci.jac.machine.jac_value import JacValue
from jaseci.jac.jsci_vm.op_codes import JsCmp


class MachineState:
    """Shared interpreter class across both sentinels and walkers"""

    def __init__(self, parent_override=None, caller=None):
        self.report = []
        self.report_status = None
        self.report_custom = None
        self.request_context = None
        self.runtime_errors = []
        self.yielded_walkers_ids = IdList(self)
        self.ignore_node_ids = IdList(self)
        self._parent_override = parent_override
        if not isinstance(self, Element) and caller:
            self._m_id = caller._m_id
            self._h = caller._h
        self._scope_stack = [None]
        self._jac_scope = None
        self._relevant_edges = []
        self._loop_ctrl = None
        self._stopped = None
        self._assign_mode = False
        self._jac_try_mode = 0  # 0 is false, +=1 for each nested try
        self._loop_limit = 10000
        self._cur_jac_ast = Ast("none")
        self._write_candidate = None
        self.inform_hook()

    def inform_hook(self):
        if hasattr(self, "_h"):
            self._h._machine = self

    def parent(self):  # parent here is always a sentinel
        if self._parent_override:
            return self._parent_override
        else:
            return Element.parent(self)

    def reset(self):
        self.report = []
        self.report_status = None
        self.report_custom = None
        self.runtime_errors = []
        self._scope_stack = [None]
        self._jac_scope = None
        self._loop_ctrl = None
        self._stopped = None

    def push_scope(self, scope: JacScope):
        self._scope_stack.append(scope)
        self._jac_scope = scope

    def pop_scope(self):
        self._scope_stack.pop()
        self._jac_scope = self._scope_stack[-1]

    def set_cur_ast(self, jac_ast):
        self._cur_jac_ast = jac_ast
        return jac_ast.kid

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.yielded_walkers_ids.obj_list():
            i.destroy()

    # Core State Management ##################
    def load_variable(self, name, assign_mode=None, jac_ast=None):
        val = self._jac_scope.get_live_var(
            name, create_mode=self._assign_mode if assign_mode is None else assign_mode
        )
        if val is None:
            self.rt_error(f"Variable not defined - {name}", jac_ast)
            self.push(JacValue(self))
        else:
            self.push(val)

    def candidate_writethrough(self):
        if self._write_candidate:
            self._write_candidate.save()
            self._write_candidate = None

    def perform_assignment(self, dest, src, jac_ast=None):
        if dest.check_assignable(jac_ast):
            dest.value = src.value
            dest.write(jac_ast)
        self.push(dest)

    def perform_copy_fields(self, dest, src, jac_ast=None):
        if dest.check_assignable(jac_ast):
            if not self.rt_check_type(dest.value, [Node, Edge], jac_ast):
                self.rt_error("Copy fields only applies to nodes and edges", jac_ast)
                self.push(dest)
            if dest.value.name != src.value.name:
                self.rt_error(
                    f"Node/edge arch {dest.value} don't match {src.value}!", jac_ast
                )
                self.push(dest)
            for i in src.value.context.keys():
                if i in dest.value.context.keys():
                    JacValue(
                        self, ctx=dest.value, name=i, value=src.value.context[i]
                    ).write(jac_ast)
        self.push(dest)

    def perform_increment(self, dest, src, op, jac_ast=None):
        if op == JsCmp.PEQ:
            dest.value = dest.value + src.value
        elif op == JsCmp.MEQ:
            dest.value = dest.value - src.value
        elif op == JsCmp.TEQ:
            dest.value = dest.value * src.value
        elif op == JsCmp.DEQ:
            dest.value = dest.value / src.value
        dest.write(jac_ast)
        self.push(dest)

    # Helper Functions ##################

    def inherit_runtime_state(self, mach):
        """Inherits runtime output state from another machine"""
        self.report += mach.report
        if mach.report_status:
            self.report_status = mach.report_status
        if mach.report_custom:
            self.report_custom = mach.report_custom
        self.runtime_errors += mach.runtime_errors

    def obj_set_to_jac_set(self, obj_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = JacSet()
        for i in obj_set:
            ret.add_obj(i)
        return ret

    def edge_to_node_jac_set(self, edge_set, location=None):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = JacSet()
        if not location:
            location = self.current_node
        for i in edge_set.obj_list():
            ret.add_obj(i.opposing_node(location))
        return ret

    def edges_filter_on_nodes(self, edge_set, node_set):
        """
        Returns nodes jac_set from edge jac_set from current node
        """
        ret = JacSet()
        for i in edge_set.obj_list():
            for j in node_set.obj_list():
                if i.jid in j.smart_edge_list:
                    ret.add_obj(i)
                    break
        return ret

    def check_builtin_action(self, func_name, jac_ast=None):
        """
        Takes reference to action attr, finds the built in function
        and returns new name used as hook by action class
        """
        if func_name not in live_actions.keys():
            self.rt_warn(f"Attempting auto-load for {func_name}", jac_ast)
            load_preconfig_actions(self._h)
        if func_name not in live_actions.keys():
            self.rt_warn(f"Builtin action unable to be loaded - {func_name}", jac_ast)
            return False
        return True

    def jac_try_exception(self, e: Exception, jac_ast):
        if isinstance(e, TryException):
            raise e
        else:
            raise TryException(self.jac_exception(e, jac_ast))

    def jac_exception(self, e: Exception, jac_ast):
        return {
            "type": type(e).__name__,
            "mod": jac_ast.loc[2],
            "msg": str(e),
            "args": e.args,
            "line": jac_ast.loc[0],
            "col": jac_ast.loc[1],
            "name": self.name if hasattr(self, "name") else "blank",
            "rule": jac_ast.name,
        }

    def rt_log_str(self, msg, jac_ast=None):
        """Generates string for screen output"""
        if jac_ast is None:
            jac_ast = self._cur_jac_ast
        name = self.name if hasattr(self, "name") else "blank"
        if jac_ast:
            msg = (
                f"{jac_ast.loc[2]}:{name} - line {jac_ast.loc[0]}, "
                + f"col {jac_ast.loc[1]} - rule {jac_ast.name} - {msg}"
            )
        else:
            msg = f"{msg}"
        return msg

    def rt_warn(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        error = self.rt_log_str(error, jac_ast)
        logger.warning(str(error))

    def rt_error(self, error, jac_ast=None):
        """Prints runtime error to screen"""
        if self._jac_try_mode:
            raise Exception(error)
        error = self.rt_log_str(error, jac_ast)
        logger.error(str(error))
        self.runtime_errors.append(error)

    def rt_info(self, msg, jac_ast=None):
        """Prints runtime info to screen"""
        logger.info(str(self.rt_log_str(msg, jac_ast)))

    def rt_check_type(self, obj, typ, jac_ast=None):
        """Prints error if type mismatach"""
        if not isinstance(typ, list):
            typ = [typ]
        for i in typ:
            if isinstance(obj, i):
                return True
        self.rt_error(
            f"Incompatible type for object "
            f"{obj} - {type(obj).__name__}, "
            f"expecting {typ}",
            jac_ast,
        )
        return False

    def get_info(self):
        return {
            "report": copy(self.report),
            "report_status": self.report_status,
            "report_custom": self.report_custom,
            "request_context": self.request_context,
            "runtime_errors": self.runtime_errors,
        }


class TryException(Exception):
    def __init__(self, ref: dict):
        super().__init__(ref["msg"])
        self.ref = ref



spawn_graph_node = """
    node test_node {has name;}
    graph test_graph {
        has anchor graph_root;
        spawn {
            graph_root =
                spawn node::test_node(name="graph_root_node_name");
        }
    }
    walker init {
        root {
            spawn here ++> graph::test_graph;
            take --> node::test_node;
        }
        test_node {
            std.out(here.name);
        }
    }
    """

basic_arith = """
    walker init {
        a = 4 + 4;
        b = 4 * -5;
        c = 4 / 4;  # Returns a floating point number
        d = 4 - 6;
        e = a + b + c + d;
        std.out(a, b, c, d, e);
    }
    """

more_arith = """
    walker init {
        a = 4 ^ 4; b = 9 % 5; std.out(a, b);
    }
    """

compare = """
    walker init {
        a = 5; b = 6;
        std.out(a == b,
                a != b,
                a < b,
                a > b,
                a <= b,
                a >= b,
                a == b-1);
    }
    """

logical = """
    walker init {
        a = true; b = false;
        std.out(a,
                !a,
                a && b,
                a || b,
                a and b,
                a or b,
                !a or b,
                !(a and b));
    }
    """

assignments = """
    walker init {
        a = 4 + 4; std.out(a);
        a += 4 + 4; std.out(a);
        a -= 4 * -5; std.out(a);
        a *= 4 / 4; std.out(a);
        a /= 4 - 6; std.out(a);

        # a := here; std.out(a);
        # Noting existence of copy assign, described later
    }
    """

if_stmt = """
    walker init {
        a = 4; b = 5;
        if(a < b): std.out("Hello!");
    }
    """

else_stmt = """
    walker init {
        a = 4; b = 5;
        if(a == b): std.out("A equals B");
        else: std.out("A is not equal to B");
    }
    """

elif_stmt = """
    walker init {
        a = 4; b = 5;
        if(a == b): std.out("A equals B");
        elif(a > b): std.out("A is greater than B");
        elif(a == b - 1): std.out("A is one less than B");
        elif(a == b - 2): std.out("A is two less than B");
        else: std.out("A is something else");
    }
    """

for_stmt = """
    walker init {
        for i=0 to i<10 by i+=1:
            std.out("Hello", i, "times!");
    }
    """
while_stmt = """
    walker init {
        i = 5;
        while(i>0) {
            std.out("Hello", i, "times!");
            i -= 1;
        }
    }
    """

break_stmt = """
    walker init {
        for i=0 to i<10 by i+=1 {
            std.out("Hello", i, "times!");
            if(i == 6): break;
        }
    }
    """

continue_stmt = """
    walker init {
        i = 5;
        while(i>0) {
            if(i == 3){
                i -= 1; continue;
            }
            std.out("Hello", i, "times!");
            i -= 1;
        }
    }
    """

continue_stmt2 = """
    walker init {
        z=10;

        while(z>=0){
            z-=1;
            if(z<5){
                continue;
            }
            std.out("hello", z);
        }
    }
    """

destroy_disconn = """
    node testnode {
        has apple;
    }

    walker init{
        node1 = spawn here ++> node::testnode ;
        node2 = spawn here ++> node::testnode ;
        node1 ++> node2;
        std.out(-->);
        destroy node1;
        # All node destroys queue'd after walk
        # may not be a good idea, must think about it
        std.out(-->);
        here !--> node2;
        std.out('1', -->);
    }
    """

array_assign = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = [[1,2],[3,4]];
            take node1;
        }
        testnode {
            a = [[0,0],[0,0]];
            std.out(a);
            a[0] = [1,1];
            std.out(a);
            std.out(here.apple);
            here.apple[1] = here.apple[0];
            std.out(here.apple);
        }
    }
    """

array_md_assign = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = [[1,2],[3,4]];
            take node1;
        }
        testnode {
            std.out(here.apple);
            here.apple[0][1] = here.apple[0][0];
            std.out(here.apple);
        }
    }
    """

dereference = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            std.out(&node1);
        }
    }
    """

pre_post_walking = """
    node testnode {
        has apple;
    }

    walker init {
        has count;

        with entry {
            count = 5;
            spawn here ++> node::testnode ;
            spawn here ++> node::testnode ;
            spawn here ++> node::testnode ;
            take -->;
        }

        testnode {
            count += 1;
        }

        with exit {std.out("count:",count);}
    }
    """

pre_post_walking_dis = """
    node testnode {
        has apple;
    }

    walker init {
        has count;

        with entry {
            count = 5;
            spawn here ++> node::testnode ;
            spawn here ++> node::testnode ;
            spawn here ++> node::testnode ;
            take -->;
        }

        testnode {
            count += 1;
            disengage;
            std.out("test");
        }

        with exit {std.out("count:",count);}
    }
    """

length = """
    node testnode {
        has apple;
    }

    walker init {
        spawn here ++> node::testnode ;
        spawn here ++> node::testnode ;
        spawn here ++> node::testnode ;
        std.out((-->).length);
        var = -->;
        std.out(var.length);
    }
    """

sort_by_col = """
    walker init {
        lst=[['b', 333],['c',245],['a', 56]];
        std.out(lst);
        std.out(std.sort_by_col(lst, 0));
        std.out(std.sort_by_col(lst, 0, true));
        std.out(std.sort_by_col(lst, 1));
        std.out(std.sort_by_col(lst, 1, true));
    }
    """

list_remove = """
    node testnode { has lst; }

    walker init {
        nd=spawn here ++> node::testnode ;
        nd.lst=[['b', 333],['c',245],['a', 56]];
        std.out(nd.lst);
        destroy nd.lst[1];
        std.out(nd.lst);
        destroy nd.lst[1];
        std.out(nd.lst);
    }
    """

can_action = """
    node testnode {
        has anchor A;
        can ptest {
            b=7;
            std.out(A,b);
            ::ppp;
        }
        can ppp {
            b=8;
            std.out(A,b);
        }
    }

    walker init {
        a= spawn here ++> node::testnode (A=56);
        a::ptest;
    }
    """

can_action_params = """
    node testnode {
        has anchor A;
        can ptest {
            b=7;
            std.out(A,b);
            ::ppp;
        }
        can ppp {
            b=8;
            std.out(A,b);
        }
    }

    walker init {
        a= spawn here ++> node::testnode (A=56);
        a::ptest(A=43);
        a::ptest(A=a.A+5);
    }
    """

cross_scope_report = """
    node testnode {
        has anchor A;
        can ptest {
            b=7;
            std.out(A,b);
            report A;
            ::ppp;
            skip;
            std.out("shouldnt show this");
        }
        can ppp {
            b=8;
            std.out(A,b);
            report b;
        }
    }

    walker init {
        a= spawn here ++> node::testnode (A=56);
        a::ptest;
        report here;
    }
    """

has_private = """
    node testnode {
        has apple;
        has private banana, grape;
    }

    walker init {
        root {
            spawn here ++> node::testnode (apple=5, banana=6, grape=1);
            take -->;
        }
        testnode {
            here.apple+=here.banana+here.grape;
            report here;
        }
    }
    """

array_idx_of_expr = """
    node testnode {
        has apple;
    }

    walker init {
        spawn here ++> node::testnode ;
        spawn here ++> node::testnode ;
        spawn here ++> node::testnode ;
        std.out((-->).length);
        var = -->[0];
        std.out([var].length);
    }
    """

dict_assign = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = {"one": 1, "two": 2};
            take node1;
        }
        testnode {
            a =  {"three": 3, "four": 4};
            std.out(a);
            a["four"] = 55;
            std.out(a);
            std.out(here.apple);
            here.apple["one"] = here.apple["two"];
            std.out(here.apple["one"]);
        }
    }
    """

dict_md_assign = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = {"one": {"inner": 44}, "two": 2};
            take node1;
        }
        testnode {
            std.out(here.apple);
            here.apple["one"]["inner"] = here.apple["two"];
            std.out(here.apple["one"]);
            std.out(here.apple["one"]['inner']);
        }
    }
    """

dict_keys = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = {"one": {"inner": 44}, "two": 2};
            take node1;
        }
        testnode {
            std.out(here.apple);
            for i in here.apple.keys:
                if(i == 'one'):
                    for j in here.apple.keys:
                        if(j == 'two'):
                            here.apple[i]["inner"] = here.apple[j];
            std.out(here.apple["one"]);
            std.out(here.apple["one"]['inner']);
        }
    }
    """
cond_dict_keys = """
    node testnode {
        has apple;
    }

    walker init{
        root {
            node1 = spawn here ++> node::testnode ;
            node1.apple = {"one": {"inner": 44}, "two": 2};
            take node1;
        }
        testnode {
            std.out(here.apple);
            if('one' in here.apple.keys) {std.out('is here');}
            if('three' not in here.apple.keys) {std.out('also not here'); }
            if('three' in here.apple.keys) {std.out('SHOULD NOT PRINT'); }
        }
    }
    """

soft_max = """
    walker init{
        can vector.softmax;
        scores = [3.0, 1.0, 0.2];
        a=vector.softmax(scores);
        report a;
        std.out(a);
    }
    """

fam_example = """
    node man;
    node woman;

    edge mom;
    edge dad;
    edge married;

    walker create_fam {
        root {
            spawn here ++> node::man;
            spawn here ++> node::woman;
            --> node::man <+[married]+> --> node::woman;
            take -->;
        }
        woman {
            son = spawn here <+[mom]+ node::man;
            son <+[dad]+ <-[married]->;
        }
        man {
            std.out("I didn't do any of the hard work.");
        }
    }
    """

visitor_preset = """
    node person {
        has name;
        has byear;
        can date.quantize_to_year::std.time_now()::>byear with setter entry;
        can std.out::byear," from ",visitor.info:: with exit;
        can std.out::byear," init only from ",visitor.info:: with init exit;
    }

    walker init {
        has year=std.time_now();
        root {
            person1 = spawn here ++>
                node::person(name="Josh", byear="1995-01-01");
            take --> ;
        }
        person {
            spawn here walker::setter;
        }
    }

    walker setter {
        has year=std.time_now();
    }
    """

visitor_local_aciton = """
    node person {
        has name;
        has byear;
        can set_year with setter entry {
            byear = visitor.year;
        }
        can print_out with exit {
            std.out(byear, " from ", visitor.info);
        }
        can reset {  # <-- Could add 'with activity' for equivalent behavior
            byear = "1995-01-01";
            std.out("resetting year to 1995:", here.context);
        }
    }

    walker init {
        has year = std.time_now();
        has person1;
        root {
            person1 = spawn here ++> node::person;
            std.out(person1.context);
            person1::reset;
            take -->;
        }
        person {
            spawn here walker::setter;
            person1::reset(name="Joe");
        }
    }

    walker setter {
        has year = std.time_now();
    }
    """

copy_assign_to_edge = """
    node person: has name, age, birthday, profession;
    edge friend: has meeting_place;
    edge family: has kind;

    walker init {
        person1 = spawn here +[friend(meeting_place = "college")]+>
            node::person(name = "Josh", age = 32);
        person2 = spawn here +[family(kind = "sister")]+>
            node::person(name = "Jane", age = 30);

        twin1 = spawn here +[friend]+> node::person;
        twin2 = spawn here +[family]+> node::person;
        twin1 := person1;
        twin2 := person2;

        -->.edge[2] := -->.edge[0];
        -->.edge[3] := -->.edge[1];

        std.out("Context for our people nodes and edges:");
        for i in -->: std.out(i.context, '\\n', i.edge[0].context);
    }
    """
